/*  Copyright (C) 1999  Demi-god / Radical Corruption
    Demi-FTPd server daemon

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#ifndef __USE_XOPEN
#define __USE_XOPEN
#endif
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <fnmatch.h>
#include <getopt.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#ifndef __USE_GNU
#define __USE_GNU
#endif
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <signal.h>
#include <dirent.h>
#include <sys/time.h>
#include <stdarg.h>
#include <dlfcn.h>
#include "dftpd.h"

/* libdftp.c dependencies */

void dftp_ufree(struct dftp_uent *);
void dftp_gfree(struct dftp_gent *);
void dftp_utmpupdate(int, struct dftp_utmpent *);
void dftp_utmpdelent(int, int);
int dftp_getline(int, char **, char **);
int dftp_readconfig(struct dftp_config *);
void dftp_freeconfig(struct dftp_config *);

/* aux.c dependencies */

void writes(const char *, ...);
void cwrites(const char *, ...);
short gethname(struct sockaddr_in *, char *);
void resetvars();

/* cmd.c dependencies */

void cmdloop();

/* srv-common.c dependencies */

int daemonize();
short openlogs(int);
int checkidentip(struct sockaddr_in *, int);
void ftplog(const char *, const char *, ...);
void readline(int, char *);
void inst_str(char *);
void reloadconf(int);
void loadplugins();

/* private variables */

extern char **environ; char **Argv; int Argc;

struct dftp_config cfg;
struct dftp_vars var;

int serv_sockfd = -1,pasv_servsockfd = -1,sockfd = -1,inetd = 0,tcpentry;
int passwdfile = -1,groupfile = -1,utmpfile = -1,dupefile = -1;
int numsitecmd = 0,*logfile = NULL;

struct sockaddr_in cli_addr, data_addr, serv_addr, pasv_addr;
struct dftp_utmpent cur_utmp;

void parent_term(int sig)
{
   close(serv_sockfd);
   ftplog("STOP","Demi-FTPd server process terminating on signal %i (%s)",sig,
    strsignal(sig));
   if(sig != SIGPIPE) exit(1);
   exit(1);
}

void parent_quit(int sig)
{
   close(serv_sockfd);
   ftplog("STOP","Demi-FTPd server process terminating on signal %i (%s)",sig,
    strsignal(sig));
   signal(SIGTERM,SIG_IGN); kill(-(getpid()), SIGTERM);
   exit(1);
}

void childexit()
{
   close(sockfd);
   if(pasv_servsockfd >= 0) close(pasv_servsockfd);
   
   if(utmpfile >= 0)
   {
      flock(utmpfile,LOCK_EX); dftp_utmpdelent(utmpfile,getpid());
      flock(utmpfile,LOCK_UN);
   }
}

void alarmhandle(int sig)
{
   if(sig != SIGALRM) return;

   ftplog("CONN","Idle timer expired. Control connection closed.");
   writes("220 Idle timer expired. Closing control connection.\r\n");
   childexit(); exit(1);
}

void activity(const char *event)
{
   int i; struct itimerval value, value2;

   cur_utmp.pid = getpid(); time(&cur_utmp.idleaika);
   if(event == NULL)
   {
      memset(cur_utmp.activity,0,ACTIVITYLEN);
      if(var.logged >= 2) value.it_value.tv_sec = cfg.maxidletime*60;
      else value.it_value.tv_sec = cfg.loginidletime*60;
      value.it_value.tv_usec = 0;
      value.it_interval.tv_sec = 0;
      value.it_interval.tv_usec = 0;
      signal(SIGALRM,alarmhandle);
      setitimer(ITIMER_REAL,&value,&value2);
      if(cfg.debuglevel >= 3) ftplog("TIMER","DEBUG: idle timer activated: set to %u mins %u secs",value.it_value.tv_sec/60,value.it_value.tv_sec%60);
   }
   else
   {
      memset(cur_utmp.activity,0,ACTIVITYLEN+1);
      strncpy(cur_utmp.activity,event,ACTIVITYLEN);
      for(i = 0; i < strlen(cur_utmp.activity); i++)
         if(!isprint(cur_utmp.activity[i]))
         {
            if(isspace(cur_utmp.activity[i])) cur_utmp.activity[i] = 0;
            else cur_utmp.activity[i] = '?';
         }
      signal(SIGALRM,SIG_IGN);
      value.it_value.tv_sec = 0;
      value.it_value.tv_usec = 0;
      value.it_interval.tv_sec = 0;
      value.it_interval.tv_usec = 0;
      setitimer(ITIMER_REAL,&value,&value2);
      if(cfg.debuglevel >= 3) ftplog("TIMER","DEBUG: idle timer deactivated, was: %u mins %u secs",value2.it_value.tv_sec/60,value2.it_value.tv_sec%60);
   }
   flock(utmpfile,LOCK_EX);
   dftp_utmpupdate(utmpfile,&cur_utmp);
   flock(utmpfile,LOCK_UN);
}

void child_term(int sig)
{
   if(sig != SIGPIPE) ftplog("STOP","Demi-FTPd child process terminating on signal %i (%s)",sig,
    strsignal(sig));
   else ftplog("CONN","Connection lost");
   exit(2);
}

void close_dataconn(int datasock)
{
   close(datasock); memset(cur_utmp.datahostname,0,HOSTNAMELEN+1);
}

int open_dataconn(char type, char state, char *str)
{
   int datasock = -1, psize;

   if(pasv_servsockfd < 0 && checkidentip(&data_addr,state))
   {
      datasock = -1;
      writes("525 Can't open data connection - ident/ip check failed\r\n");
      if(cfg.debuglevel >= 2) ftplog("AUTH","Cannot open data connection - ident/ip check failed");
      return -1;   
   }
   writes("150 Opening ");
   switch(type)
   {
      case 'A': writes("ASCII"); break;
      case 'I': writes("BINARY"); break;
   }
   writes(" mode data connection for %s.\r\n",str);
   if(pasv_servsockfd < 0)
   {
      if( (datasock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
      {
         writes("425 Can't open stream socket.\r\n");
         ftplog("ERR","Cannot open stream socket"); return -1;
      }
      if(connect(datasock, (struct sockaddr *) &data_addr, sizeof(data_addr)) < 0)
      {
         close(datasock);
         writes("425 Can't open data connection.\r\n");
         ftplog("ERR","Cannot open data connection"); return -1;
      }
      gethname(&data_addr,cur_utmp.datahostname);
   }
   else
   {
      psize = sizeof(pasv_addr);
      datasock = accept(pasv_servsockfd, (struct sockaddr *) &pasv_addr, &psize);
      close(pasv_servsockfd); pasv_servsockfd = -1;
      if(datasock < 0)
      {
         writes("425 Can't open passive connection.\r\n");
         return -1;
      }
      if(checkidentip(&pasv_addr,state))
      {
         close_dataconn(datasock); datasock = -1;
         writes("525 Connection rejected - ident/ip check failed\r\n");
         if(cfg.debuglevel >= 2) ftplog("AUTH","Cannot open data connection -
          ident/ip check failed"); return -1;
      }
      gethname(&pasv_addr,cur_utmp.datahostname);
   }
   return datasock;
}

int serve()
{
   char filename[PATH_MAX];

   /* Reroute signals to our own handler */

   signal(SIGCLD,SIG_IGN);
   signal(SIGALRM,SIG_IGN);
   signal(SIGHUP,child_term);
   signal(SIGTERM,child_term);
   signal(SIGQUIT,child_term);
   signal(SIGPIPE,child_term);
   signal(SIGSEGV,child_term);
   signal(SIGILL,child_term);
   signal(SIGSTKFLT,child_term);
   signal(SIGINT,child_term);
   signal(SIGFPE,child_term);
   signal(SIGTRAP,child_term);
   signal(SIGIOT,child_term);
   signal(SIGIO,child_term);
   signal(SIGBUS,child_term);
   signal(SIGUSR1,SIG_IGN);
   signal(SIGUSR2,SIG_IGN);

   /* Open user/group databases, UTMP etc. */

   sprintf(filename,"%s/%s",cfg.confdir,USERFILE);
   passwdfile = open(filename,O_RDWR);
   if(passwdfile < 0)
   {
      writes("551-Cannot open user database.\r\n");
      writes("551 Closing connection.\r\n");
      close(sockfd);
      ftplog("ERR","Failed to open user database(%s)",filename);
      return(2);
   }
   sprintf(filename,"%s/%s",cfg.confdir,GROUPFILE);
   groupfile = open(filename,O_RDWR);
   if(groupfile < 0)
   {
      writes("551-Cannot open group database.\r\n");
      writes("551 Closing connection.\r\n");
      close(sockfd);
      ftplog("ERR","Failed to open group database");
      close(passwdfile); return(2);
   }
   sprintf(filename,"%s/%s",cfg.confdir,UTMPFILE);
   utmpfile = open(filename,O_RDWR|O_CREAT,0640);
   if(utmpfile < 0)
   {
      writes("551-Cannot create/open UTMP file.\r\n");
      writes("551 Closing connection.\r\n");
      close(sockfd);
      ftplog("ERR","Failed to create/open UTMP file");
      return(2);
   }
   sprintf(filename,"%s/%s",cfg.confdir,DUPEFILE);
   dupefile = open(filename,O_RDWR|O_CREAT,0640);
   if(dupefile < 0)
   {
      writes("551-Cannot create/open dupechecker database.\r\n");
      writes("551 Closing connection.\r\n");
      close(sockfd);
      ftplog("ERR","Failed to create/open dupechecker database");
      return(2);
   }

   atexit(childexit); gethname(&cli_addr,cur_utmp.controlhostname);
   loadplugins();

   ftplog("CONN","Connection accepted from %s",cur_utmp.controlhostname);
   time(&cur_utmp.aika); activity(NULL);
   if(cfg.mod_proctitle) inst_str("Demi-FTPd - child");
   if(cfg.welcomemsg == NULL)
      writes("220 Demi-FTP server v%s ready.\r\n",VER);
   else cwrites("220 %s\r\n",cfg.welcomemsg);

   cmdloop(); return 0;
}

int main(int argc, char *argv[])
{
   int clilen,childpid,option,quiet = 0;

   if(geteuid() != 0)
   {
      fprintf(stderr,"Sorry -- root access is required.\n");
      return(1);
   }

   dftp_readconfig(&cfg);

   while( (option = getopt(argc,argv,"qip:d:")) != EOF)
   {
      if(option == 'q') quiet = -1;
      if(option == 'i') inetd = -1;
      if(option == 'd') cfg.debuglevel = atoi(optarg);
      if(option == 'p') cfg.port = atoi(optarg);
      if(option == '?') return 1;
   }

   if(!quiet && !inetd) printf("Demi-ftp FTP server v%s by Demi-god\n",VER);
   Argv = argv; Argc = argc;
   tzset(); resetvars();
   if(openlogs(quiet|inetd)) return 2;
   memset(&cur_utmp,0,sizeof(struct dftp_utmpent));
   if(inetd)
   {
      quiet = sizeof(cli_addr);
      getpeername(1,&cli_addr,&quiet);
      sockfd = 0; exit(serve());
   }

   if(!quiet)
   {
      if(cfg.debuglevel) printf("Debug level = %i\n",cfg.debuglevel);
      printf("Configuration path = \"%s\"\n",cfg.confdir);
      printf("Server port = %i\n",cfg.port);
      printf("Site directory = \"%s\"\n",cfg.sitedir);
      printf("Welcome file = \"%s\"\n",cfg.welcomefile);
      printf("Create mode = %o\n",cfg.cmode);
      for(quiet = 0; quiet < cfg.logcount; quiet++)
      {
         if(cfg.logs[quiet].path)
            printf("Log %s to %s\n",cfg.logs[quiet].pattern,cfg.logs[quiet].path);
         else printf("Do not log %s\n",cfg.logs[quiet].pattern);
      }
      for(quiet = 0; quiet < cfg.uccount; quiet++)
      {
         if(quiet == 0) printf("Upload checker\n--------------\n");
         if(*cfg.ulchecker[quiet].handler == '-')
            printf("Ban '%s'\n",cfg.ulchecker[quiet].mask);
         else if(*cfg.ulchecker[quiet].handler == '+')
            printf("Auto-accept '%s'\n",cfg.ulchecker[quiet].mask);
         else printf("Check '%s' with '%s'\n",cfg.ulchecker[quiet].mask,cfg.ulchecker[quiet].handler);
      }
      if(quiet > 0) printf("--------------\n");
      for(quiet = 0; quiet < cfg.dccount; quiet++)
      {
         if(quiet == 0) printf("Duplicate file checker\n---------------------\n");
         printf("Check '%s'\n",cfg.dupechecker[quiet]);
      }
      if(quiet > 0) printf("---------------------\n");
      if(cfg.welcomemsg != NULL) printf("Welcome message = \"%s\"\n",cfg.welcomemsg);
      if(cfg.goodbyemsg != NULL) printf("Goodbye message = \"%s\"\n",cfg.goodbyemsg);
      if(cfg.statusline != NULL) printf("Status line = \"%s\"\n",cfg.statusline);
      quiet = 0;
   }
   if(cfg.mod_proctitle) inst_str("Demi-FTPd - parent");

   if(signal(SIGTERM,parent_term) == SIG_ERR) 
      fprintf(stderr,"Warning: cannot set signal handler for SIGTERM\n");
   if(signal(SIGHUP,reloadconf) == SIG_ERR)
      fprintf(stderr,"Warning: cannot set signal handler for SIGHUP\n");
   if(signal(SIGQUIT,parent_quit) == SIG_ERR)
      fprintf(stderr,"Warning: cannot set signal handler for SIGQUIT\n");

   if(signal(SIGCLD,SIG_IGN) == SIG_ERR)
   {
      fprintf(stderr,"Fatal: could not ignore SIGCLD signal\n");
      close(serv_sockfd); return 8;
   }
   ftplog("START","Demi-FTPd v%s starting up",VER);

   memset(&serv_addr,0,sizeof(serv_addr));
   serv_addr.sin_family = AF_INET;
   serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
   serv_addr.sin_port = htons(cfg.port);
   if( (serv_sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
   {
      fprintf(stderr,"Can't open stream socket\n");
      return 3;
   }
   option = 1;
   if(setsockopt(serv_sockfd,SOL_SOCKET,SO_REUSEADDR,&option,sizeof(option)))
   {
      fprintf(stderr,"Cannot set socket options: %s\n",strerror(errno));
      return 3;
   }
   if(!quiet)
   {
      printf("Binding to port %hu...",cfg.port); fflush(stdout);
   }
   if(bind(serv_sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
   {
      fprintf(stderr,"Can't bind local address(%s)\n",strerror(errno));
      return 4;
   }
   if(!quiet) printf("done\n");

   listen(serv_sockfd,BACKLOG);
   if(daemonize()) return(5);

   for(;;)
   {
      clilen = sizeof(cli_addr);
      sockfd = accept(serv_sockfd, (struct sockaddr *) &cli_addr, &clilen);
      if(sockfd < 0)
      {
          ftplog("ERR","accept() failed: %s",strerror(errno)); return 6;
      }

      if( (childpid = fork()) < 0)
      {
          ftplog("ERR","fork() failed: %s",strerror(errno)); return 7;
      }
      else if(childpid == 0)
      {
         close(serv_sockfd);
         exit(serve());
      }
      close(sockfd);
   }
}
