#ifndef __DFTPD_H__
#define __DFTPD_H__

#include <time.h>

#define USERFILE "passwd.dftp"
#define GROUPFILE "group.dftp"
#define UTMPFILE "utmp.dftp"
#define DUPEFILE "dupebase.dftp"
#define PLUGINDIR "plugins"
#define DFTP_ENV "DFTPD_CONFDIR"
#define CONFIGFILE "dftpd.conf"

static const char conf_defaultpath[] = "/etc/dftpd";
static const char site_defaultpath[] = "/site";
static const char bin_defaultpath[] = "/usr/local/dftpd";

#define FIRSTUID 100
#define FIRSTGID 100
#define DEFAULTPORT 5444
#define IDENTPORT 113
#define BACKLOG 5
#define MAXLINE 512
#define NAMELEN 16
#define HOSTNAMELEN 40
#define ACTIVITYLEN 60
#define PASSWDLEN 13

#define KB 1024
#define MB 1048576
#define GB 1073741824

#define GUEST 1
#define ANONYMOUS 2
#define REAL 4
#define ALL 7

#define LS_LONG			1
#define LS_REVERSE		2
#define LS_RECURSIVE	4
#define LS_NORESOLVE	8
#define LS_ALMOSTALL	16

#define NOSPACES 0
#define SPACES 1

#define ID_CONTROL 1
#define ID_DATA    2
#define ID_BAN     4

#define VER "1.3"

#define UTMPSTRING "%-16.16s:%7.7i:%9.9li:%9.9li:%-60.60s:%-40.40s:%-40.40s\n"

struct dftp_log
{
   char *pattern;
   char *path;
};

struct dftp_ulchecker
{
   char *mask;
   char *handler;
};

struct dftp_flag
{
   char mark;
   char *name;
};

struct dftp_sitecmd
{
   char *name;
   char *helptxt;
   char avail;
   void (*cmd)(char *);
};

struct dftp_config
{
   char *confdir;
   char *sitedir;
   struct dftp_log *logs; int logcount;
   struct dftp_sitecmd *sitecmds; int sitecmdcount;
   struct dftp_flag *flags; int flagcount;
   unsigned short port, httpport;
   char debuglevel;
   char maxloginfails;
   int cmode;
   unsigned short maxidletime, loginidletime;
   unsigned int maxviewsize;
   unsigned int stupdatetime;
   struct dftp_ulchecker *ulchecker; int uccount;
   char **dupechecker; int dccount;
   char *welcomefile;
   char *welcomemsg, *goodbyemsg;
   char *statusline;
   char mod_proctitle;
   char resolvehosts;
   char realusersupport;
   char shadow;
};

struct ident_ip
{
   unsigned char *ident;
   unsigned char *ip;
   unsigned char flags;
};

struct dftp_uent
{
   char name[NAMELEN+1];
   char passwd[PASSWDLEN+1];
   unsigned char *flags;
   unsigned short maxconn;
   unsigned int uid;
   unsigned int gid;
   unsigned short ratio;
   unsigned long long ulbytes;
   unsigned long long dlbytes;
   unsigned long long credits;
   char *homedir;
   char *comment;
   int identipcount;
   struct ident_ip *identip;
};

struct dftp_gent
{
   char name[NAMELEN+1];
   unsigned short id;
   int membercount;
   char **members;
};

struct dftp_utmpent
{
   char name[NAMELEN+1];
   int pid;
   time_t aika, idleaika;
   char activity[ACTIVITYLEN+1];
   char controlhostname[HOSTNAMELEN+1];
   char datahostname[HOSTNAMELEN+1];
};

struct dftp_dupeent
{
   char *filename;
   time_t ultime;
   char uploader[NAMELEN+1];
   char *dirname;
};

struct dftp_vars
{
   short logged;
   char *ident, identfetched;
   struct dftp_uent user;
   struct dftp_gent group;
   char trans_type;
   int cmode;
   int userclass;
   unsigned short loginfails;
   unsigned long long sessionul;
   unsigned long long sessiondl;
   long offset;
};

struct dftp_plugininfo
{
   struct dftp_vars *var;
   struct dftp_config *cfg;
   void (*writes)(const char *,...);
   void (*cwrites)(const char *,...);
   void (*ftplog)(const char *,const char *,...);
   void (*activity)(const char *);
   void (*viewfile)(const char *, const short);
   int passwdfile, groupfile, utmpfile, sockfd;
   char ftpdversion[10];
};
#endif
