/*  Copyright (C) 1999  Demi-god / Radical Corruption
    HTTP Remote Configuration Server for Demi-FTPd

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#ifndef __USE_XOPEN
#define __USE_XOPEN
#endif
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#ifndef __USE_GNU
#define __USE_GNU
#endif
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <ctype.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "dftpd.h"

/* libdftp.c dependencies */

int dftp_readconfig(struct dftp_config *);
void dftp_freeconfig(struct dftp_config *cnf);
int dftp_utmpgetent(int, struct dftp_utmpent *);
void dftp_uaddent(int, struct dftp_uent *);
int dftp_ugetent(int, struct dftp_uent *);
int dftp_ugetbyname(int, char *, struct dftp_uent *);
int dftp_ugetbyid(int, int, struct dftp_uent *);
void dftp_uupdatedb(int, struct dftp_uent *);
void dftp_uupdatesitedb(int, int);
void dftp_ufree(struct dftp_uent *);
int dftp_udelent(int, unsigned short);
void dftp_gaddent(int, struct dftp_gent *);
int dftp_ggetent(int, struct dftp_gent *);
int dftp_ggetbyid(int, int, struct dftp_gent *);
void dftp_gupdatedb(int, struct dftp_gent *);
void dftp_gupdatesitedb(int, int);
void dftp_gfree(struct dftp_gent *);
int dftp_gdelent(int, unsigned short);
char *dftp_identip2text(struct ident_ip *, int);
char *dftp_printtime(time_t);
char *dftp_printbytes(unsigned long long);
void dftp_creds(char *, unsigned long long *);

/* srv-common.c dependencies */

int daemonize();
short openlogs(int);
int checkidentip(struct sockaddr_in *, int);
void writes(const char *format, ...);
void ftplog(const char *, const char *, ...);
void readline(int, char *);
void inst_str(char *);
void reloadconf(int);
void loadplugins();

/* private variables */

const char _headers_[] = "Content-Type: text/html; charset=iso-8859-1\r\n\r\n";
const char _htmltype_[] = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\">\r\n";
const char invalidchars[] = ": \"%&/()=\\";

char **Argv, *password = NULL; int Argc;
int serv_sockfd, sockfd, inetd = 0;
int passwdfile = -1, utmpfile = -1, groupfile = -1, *logfile = NULL;
int passwdsitefile = -1, groupsitefile = -1;
struct dftp_config cfg;
struct dftp_uent user;
struct sockaddr_in cli_addr, serv_addr;

/* Signal handlers */

void httplostconn(int sig)
{
   if(cfg.debuglevel > 1) ftplog("CONN","HTTP: Connection unexpectedly closed");
   exit(1);
}

void httpquit(int sig)
{
   ftplog("STOP","HTTP: Server shutting down"); exit(0);
}

void httperr(int sig)
{
   ftplog("ERR","HTTP: Server terminating on signal %i(%s)",sig,strsignal(sig));
   exit(1);
}

void httpclienterr(int sig)
{
   ftplog("ERR","HTTP: Abnormal termination on signal %i(%s)",sig,strsignal(sig));
   exit(1);
}

/* Utility functions */

void cmd_err(const char *error, const char *errtxt, const char *meta)
{
   writes("HTTP/1.0 %s\r\n",error);
   if(meta) writes("%s\r\n",meta);
   writes(_headers_);
   writes(_htmltype_);
   writes("<HTML><HEAD><TITLE>%s</TITLE></HEAD>\r\n",error);
   writes("<BODY STYLE=\"background-color: silver;\"><H1>%s</H1></BODY></HTML>\r\n",errtxt);
   close(sockfd); exit(1);
}

char *parseargs(char *args)
{
   char *line, *lptr, *ptr = args; unsigned int ccode = 0;

   line = malloc(strlen(args)+1); lptr = line; strcpy(line,args);
   while(*ptr)
   {
      if(*ptr == '+') *lptr = ' ';
      else if(*ptr == '%')
      {
         sscanf(ptr+1,"%2X",&ccode);
         *lptr = (char)ccode; ptr += 2;
      }
      else *lptr = *ptr;
      ptr++; lptr++;
   }
   *lptr = 0;
   return line;
}

int httpopencfgfile(int *fd, const char *name, int mode)
{
   char path[PATH_MAX];

   if(*fd >= 0) return -1;
   sprintf(path,"%s/%s",cfg.confdir,name);
   *fd = open(path,mode);
   if(*fd < 0)
   {
      if(cfg.debuglevel > 0) ftplog("ERR","HTTP: Cannot open %s",path);
      return 0;
   }
   return -1;
}

int httpopensitefile(int *fd, const char *name, int mode)
{
   char path[PATH_MAX];

   if(*fd >= 0) return -1;
   sprintf(path,"%s/etc/%s",cfg.sitedir,name);
   *fd = open(path,mode|O_CREAT,0644);
   if(*fd < 0)
   {
      ftplog("ERR","HTTP: Cannot open %s",path);
      return 0;
   }
   return -1;
}

void monitor(char *args)
{
   char *token, *value; struct dftp_utmpent entry; int result, refreshtime = 0, ulevel = 0, killpid = 0;
   
   if(user.flags && strchr(user.flags,'S')) ulevel = 2;
   else if(user.flags && strchr(user.flags,'X')) ulevel = 1;
   writes("HTTP/1.0 200 OK\r\n");
   while(args && *args)
   {
      value = strsep(&args,"&"); if(!value) value = args;
      if(strchr(value,'='))
      {
         token = strsep(&value,"=");
         if(!strcmp(token,"refresh")) refreshtime = atoi(value);
         else if(!strncmp(token,"kick-",5)) killpid = atoi(token+5);
      }
   }
   if(refreshtime) writes("Refresh: %i\r\n",refreshtime);
   writes(_headers_);
   writes(_htmltype_);
   writes("<HTML><HEAD><TITLE>Demi-FTPd monitor</TITLE></HEAD>\r\n");
   writes("<BODY STYLE=\"text-align: center; background-color: silver;\">\r\n");
   writes("<FORM ACTION=\"monitor\" METHOD=\"POST\">\r\n");
   writes("<TABLE BORDER>\r\n");
   writes("<TR><TH><B>User name</B></TH>\r\n");
   if(ulevel) writes("<TH><B>Pid</B></TH>\r\n");
   writes("<TH><B>Online time</B></TH>\r\n");
   writes("<TH><B>Activity</B></TH>\r\n");
   if(ulevel) writes("<TH><B>Host name</B></TH>\r\n");
   if(ulevel) writes("<TH><B>Data connection host name</B></TH>\r\n");
   if(ulevel > 1) writes("<TH></TH>\r\n");
   writes("</TR>\r\n");
   if(httpopencfgfile(&utmpfile,UTMPFILE,O_RDONLY))
   {
      flock(utmpfile,LOCK_EX); lseek(utmpfile,0,SEEK_SET);
      while((result = dftp_utmpgetent(utmpfile,&entry)))
         if(result < 0)
         {
            if(killpid == entry.pid && (ulevel > 1 || !strncmp(user.name,entry.name,strlen(user.name))))
            {
               kill(killpid,SIGTERM); ftplog("CMD","HTTP: killed pid %hu",killpid);
            }
            writes("<TR><TD ALIGN=CENTER>%s</TD>",entry.name);
            if(ulevel) writes("<TD ALIGN=CENTER>%hu</TD>",entry.pid);
            writes("<TD ALIGN=CENTER>%s</TD>",dftp_printtime(entry.aika));
            if(entry.activity[0] == 32) writes("<TD ALIGN=CENTER>Idle %s</TD>",dftp_printtime(entry.idleaika));
            else writes("<TD>%s</TD>",entry.activity);
            if(ulevel) writes("<TD ALIGN=CENTER>%s</TD>",entry.controlhostname);
            if(ulevel) writes("<TD ALIGN=CENTER>%s</TD>",entry.datahostname);
            if(ulevel > 1 || !strncmp(user.name,entry.name,strlen(user.name)))
               writes("<TD><INPUT TYPE=\"submit\" NAME=\"kick-%hu\" VALUE=\"Kick\"></TH>",entry.pid);
            writes("</TR>\r\n");
         }
      flock(utmpfile,LOCK_UN);
   }
   writes("</TABLE><P></FORM>\r\n");
   writes("<FORM METHOD=\"GET\" ACTION=\"/monitor\"><HR>\r\n");
   writes("<P STYLE=\"margin-top: 0;\">\r\n");
   writes("<A HREF=\"/\">Back</A> \r\n");
   writes("Auto-refresh: ");
   writes("<INPUT TYPE=\"text\" NAME=\"refresh\" SIZE=3");
   if(refreshtime) writes(" VALUE=\"%u\"",refreshtime);
   writes("> seconds <INPUT TYPE=\"submit\" VALUE=\"Engage\"></P></FORM>\r\n");
   writes("</BODY></HTML>\r\n");
}

/* User information handling related routines */

unsigned int getnextuid()
{
   unsigned int id = FIRSTUID, *uids = NULL; unsigned int uidcount = 0; int i;
   struct dftp_uent entry;

   if(!httpopencfgfile(&passwdfile,USERFILE,O_RDWR)) return id;
   flock(passwdfile,LOCK_EX); lseek(passwdfile,0,SEEK_SET);
   while(dftp_ugetent(passwdfile,&entry))
   {
      uids = realloc(uids,++uidcount*sizeof(unsigned int));
      uids[uidcount-1] = entry.uid;
      dftp_ufree(&entry);
   }
   flock(passwdfile,LOCK_UN);
   for(i=0;i<uidcount;i++)
      if(id == uids[i])
      {
         id++; i = -1;
      }
   return id;
}

unsigned int getfirstgid()
{
   int i; struct dftp_gent entry;

   if(!httpopencfgfile(&groupfile,GROUPFILE,O_RDONLY)) return 0;
   flock(groupfile,LOCK_EX); lseek(groupfile,0,SEEK_SET);
   while((i=dftp_ggetent(groupfile,&entry)) == 1);
   flock(groupfile,LOCK_UN);
   if(i != -1) return 0;
   return entry.id;
}

void chpassword(unsigned int uid)
{
   writes("HTTP/1.0 200 OK\r\n");
   writes(_headers_);
   writes(_htmltype_);
   writes("<HTML><HEAD><TITLE>Demi-FTPd User Configuration</TITLE></HEAD>\r\n");
   writes("<BODY STYLE=\"text-align: center; font-size: 18pt; background-color: silver;\">\r\n");
   writes("<FORM METHOD=\"POST\" ACTION=\"userconfig\"><PRE>\r\n");
   writes("<INPUT TYPE=\"hidden\" NAME=\"user\" VALUE=\"%hu\">\r\n",uid);
   writes("Password <INPUT TYPE=\"password\" NAME=\"pass1\"><BR>\r\n");
   writes("   Again <INPUT TYPE=\"password\" NAME=\"pass2\"><BR>\r\n");
   writes("<A HREF=\"userconfig\">Back</A> <INPUT TYPE=\"submit\" VALUE=\"Change\">\r\n");
   writes("</PRE></FORM></BODY></HTML>\r\n"); exit(0);
}

void usrlist(char *);

void getuser(char *args, struct dftp_uent *entry)
{
   int result, newpass = 0; char *token, *value, *pass = NULL; unsigned short unum = 0;

   args = parseargs(args);
   value = strsep(&args,"&"); if(!value) value = args;
   token = strsep(&value,"=");
   if(token && !strcmp(token,"user")) unum = atoi(value);
   if(!args || !unum)
   {
      if(!unum)
      {
         if(!(entry->gid = getfirstgid()))
            cmd_err("404 Not found","No groups found - create a group before adding users",NULL);
         entry->uid = getnextuid(); sprintf(entry->name,"user%hu",entry->uid-FIRSTUID+1);
         entry->maxconn = 1; sprintf(entry->passwd,"*");
         entry->homedir = malloc(2); strcpy(entry->homedir,"/");
         if(!httpopensitefile(&passwdsitefile,"passwd",O_RDWR))
            cmd_err("500 Internal server error","Cannot access in-site user database",NULL);
         flock(passwdfile,LOCK_EX); lseek(passwdfile,0,SEEK_END);
         dftp_uaddent(passwdfile,entry); flock(passwdsitefile,LOCK_EX);
         dftp_uupdatesitedb(passwdfile,passwdsitefile);
         flock(passwdfile,LOCK_UN); flock(passwdsitefile,LOCK_UN); return;
      }
      if(!httpopencfgfile(&passwdfile,USERFILE,O_RDONLY))
         cmd_err("500 Internal server error","Cannot complete your request.",NULL);
      flock(passwdfile,LOCK_EX); lseek(passwdfile,0,SEEK_SET);
      result = dftp_ugetbyid(passwdfile,unum,entry);
      flock(passwdfile,LOCK_UN);
      if(!result) cmd_err("404 Not found","Requested user not found",NULL);
      return;
   }
   entry->uid = unum;
   if(!httpopencfgfile(&passwdfile,USERFILE,O_RDWR))
      cmd_err("500 Internal server error","Cannot access user database",NULL);
   if(!httpopensitefile(&passwdsitefile,"passwd",O_RDWR))
      cmd_err("500 Internal server error","Cannot access in-site user database",NULL);
   while(args && *args)
   {
      value = strsep(&args,"&"); if(!value) value = args;
      if(strchr(value,'='))
      {
         token = strsep(&value,"=");
         if(!strcmp(token,"name")) strncpy(entry->name,value,NAMELEN);
         else if(!strcmp(token,"pass")) strncpy(entry->passwd,value,PASSWDLEN);
         else if(!strcmp(token,"gid")) entry->gid = atoi(value);
         else if(!strcmp(token,"maxconn")) entry->maxconn = atoi(value);
         else if(!strcmp(token,"ratio")) entry->ratio = atoi(value);
         else if(!strcmp(token,"dlbytes")) dftp_creds(value,&entry->dlbytes);
         else if(!strcmp(token,"ulbytes")) dftp_creds(value,&entry->ulbytes);
         else if(!strcmp(token,"credits")) dftp_creds(value,&entry->credits);
         else if(!strcmp(token,"homedir"))
         {
            entry->homedir = malloc(strlen(value)+1); strcpy(entry->homedir,value);
         }
         else if(!strcmp(token,"comment"))
         {
            entry->comment = malloc(strlen(value)+1); strcpy(entry->comment,value);
         }
         else if(!strncmp(token,"flag-",5))
         {
            if(entry->flags)
               entry->flags = realloc(entry->flags,strlen(entry->flags)+2);
            else entry->flags = calloc(1,2);
            strcat(entry->flags,token+5);
         }
         else if(!strcmp(token,"ident"))
         {
            entry->identip = realloc(entry->identip,++entry->identipcount*sizeof(struct ident_ip));
            if(strlen(value))
            {
               entry->identip[entry->identipcount-1].ident = malloc(strlen(value)+1);
               strcpy(entry->identip[entry->identipcount-1].ident,value);
            }
            else entry->identip[entry->identipcount-1].ident = NULL;
            entry->identip[entry->identipcount-1].flags = 0;
         }
         else if(!strcmp(token,"ip"))
         {
            entry->identip[entry->identipcount-1].ip = malloc(strlen(value)+1);
            strcpy(entry->identip[entry->identipcount-1].ip,value);
         }
         else if(!strcmp(token,"cflag")) entry->identip[entry->identipcount-1].flags |= ID_CONTROL;
         else if(!strcmp(token,"dflag")) entry->identip[entry->identipcount-1].flags |= ID_DATA;
         else if(!strcmp(token,"bflag")) entry->identip[entry->identipcount-1].flags |= ID_BAN;
         else if(!strcmp(token,"delidentip"))
         {
            /* Workaround for realloc() bug found in glibc */
            if(--entry->identipcount)
               entry->identip = realloc(entry->identip,entry->identipcount*sizeof(struct ident_ip));
            else
            {
               free(entry->identip); entry->identip = NULL;
            }
         }
         else if(!strcmp(token,"addidentip"))
         {
            entry->identip = realloc(entry->identip,++entry->identipcount*sizeof(struct ident_ip));
            memset(&entry->identip[entry->identipcount-1],0,sizeof(struct ident_ip));
            entry->identip[entry->identipcount-1].ip = malloc(8);
            sprintf(entry->identip[entry->identipcount-1].ip,"0.0.0.0");
         }
         else if(!strcmp(token,"deluser"))
         {
            flock(passwdfile,LOCK_EX); result = dftp_udelent(passwdfile,unum);
            if(!result) cmd_err("404 Not found","Target user not found",NULL);
            flock(passwdsitefile,LOCK_EX); dftp_uupdatesitedb(passwdfile,passwdsitefile);
            flock(passwdfile,LOCK_UN); flock(passwdsitefile,LOCK_UN);
            usrlist(NULL); exit(0);
         }
         else if(!strcmp(token,"chpass")) newpass = 1;
         else if(!strcmp(token,"pass1"))
         {
            pass = malloc(strlen(value)+1); strcpy(pass,value);
         }
         else if(!strcmp(token,"pass2"))
         {
            time_t tme;

            if(!pass || strcmp(pass,value)) cmd_err("409 Conflict","Passwords do not match",NULL);
            flock(passwdfile,LOCK_EX); lseek(passwdfile,0,SEEK_SET);
            if(!dftp_ugetbyid(passwdfile,unum,entry))
               cmd_err("404 Not found","User not found",NULL);
            flock(passwdfile,LOCK_UN);
            time(&tme); srandom((unsigned int)tme);
            do
            {
               entry->passwd[0] = random();
            } while( !isalnum(entry->passwd[0]) );
            tme += 3; srandom((unsigned int)tme);
            do
            {
               entry->passwd[1] = random();
            } while( !isalnum(entry->passwd[1]) );
            strcpy(entry->passwd,crypt(pass,entry->passwd)); break;
         }
      }
   }
   if(strcspn(entry->name,invalidchars) < strlen(entry->name))
      cmd_err("409 Conflict","Invalid character(s) in name",NULL);
   flock(passwdfile,LOCK_EX); lseek(passwdfile,0,SEEK_SET);
   dftp_uupdatedb(passwdfile,entry);
   flock(passwdsitefile,LOCK_EX); dftp_uupdatesitedb(passwdfile,passwdsitefile);
   flock(passwdfile,LOCK_UN); flock(passwdsitefile,LOCK_UN);
   if(newpass) chpassword(unum);
}

void usredit(char *args)
{
   struct dftp_uent entry; struct dftp_gent gentry;
   int i;

   if(!httpopencfgfile(&groupfile,GROUPFILE,O_RDWR))
      cmd_err("500 Internal server error","Cannot complete your request.",NULL);
   memset(&entry,0,sizeof(entry)); if(args) getuser(args, &entry);
   writes("HTTP/1.0 200 OK\r\n");
   writes(_headers_);
   writes(_htmltype_);
   writes("<HTML><HEAD><TITLE>Demi-FTPd User Configuration</TITLE></HEAD>\r\n");
   writes("<BODY STYLE=\"text-align: center; font-size: 18pt; background-color: silver;\">\r\n");
   writes("<FORM METHOD=\"POST\" ACTION=\"userconfig\"><P>\r\n");
   writes("<INPUT TYPE=\"hidden\" NAME=\"user\" VALUE=\"%hu\">\r\n",entry.uid);
   writes("<INPUT TYPE=\"hidden\" NAME=\"pass\" VALUE=\"%s\">\r\n",entry.passwd);
   if(strcmp(entry.name,"ftp"))
      writes("Name <INPUT TYPE=\"text\" NAME=\"name\" MAXLENGTH=\"%i\" SIZE=\"%i\" VALUE=\"%s\"> ",NAMELEN,NAMELEN,entry.name);
   else writes("<B>Anonymous user</B><INPUT TYPE=\"hidden\" NAME=\"name\" VALUE=\"ftp\"> ");
   writes("Group <SELECT NAME=\"gid\">\r\n");
   flock(groupfile,LOCK_EX); lseek(groupfile,0,SEEK_SET);
   while(dftp_ggetent(groupfile,&gentry))
   {
      if(strcmp(entry.name,"ftp") || !strcmp(gentry.name,"ftp"))
      {
         writes("<OPTION VALUE=\"%hu\"",gentry.id);
         if(gentry.id == entry.gid) writes(" SELECTED");
         writes(">%s\r\n",gentry.name);
      }
   }
   flock(groupfile,LOCK_UN);
   writes("</SELECT><BR>\r\n");
   writes("<SPAN STYLE=\"font-size: smaller;\">ID=%hu</SPAN><BR>\r\n",entry.uid);
   writes(" Maximum connections <INPUT TYPE=\"text\" NAME=\"maxconn\" MAXLENGTH=2 SIZE=2 VALUE=\"%i\"> ",entry.maxconn);
   writes("Ratio <B>1:</B><INPUT TYPE=\"text\" NAME=\"ratio\" MAXLENGTH=1 SIZE=2 VALUE=\"%i\"><BR>\r\n",entry.ratio);
   writes("Home directory <INPUT TYPE=\"text\" NAME=\"homedir\" SIZE=30");
   if(entry.homedir) writes(" VALUE=\"%s\"",entry.homedir);
   writes("><BR>\r\nUser info line <INPUT TYPE=\"text\" NAME=\"comment\" SIZE=30");
   if(entry.comment) writes(" VALUE=\"%s\"",entry.comment);
   writes("></P>\r\n");
   writes("<TABLE BORDER>\r\n");
   writes("<TR><TD>Downloaded bytes</TD><TD>\r\n");
   writes("<INPUT TYPE=\"text\" NAME=\"dlbytes\" SIZE=\"12\" VALUE=\"%qu\">(%s)</TD></TR>\r\n",entry.dlbytes,dftp_printbytes(entry.dlbytes));
   writes("<TR><TD>Uploaded bytes</TD><TD>\r\n");
   writes("<INPUT TYPE=\"text\" NAME=\"ulbytes\" SIZE=\"12\" VALUE=\"%qu\">(%s)</TD></TR>\r\n",entry.ulbytes,dftp_printbytes(entry.ulbytes));
   writes("<TR><TD>Credits</TD><TD>\r\n");
   writes("<INPUT TYPE=\"text\" NAME=\"credits\" SIZE=\"12\" VALUE=\"%qu\">(%s)</TD></TR>\r\n",entry.credits,dftp_printbytes(entry.credits));
   writes("</TABLE><P STYLE=\"margin-top: 0;\"><BR><TABLE BORDER>\r\n");
   writes("<CAPTION>Flags</CAPTION>\r\n");
   for(i=0;i<cfg.flagcount;i++)
   {
      writes("<TR><TD>%s</TD><TD><INPUT TYPE=\"checkbox\" NAME=\"flag-%c\"",cfg.flags[i].name,cfg.flags[i].mark);
      if(entry.flags && strchr(entry.flags,cfg.flags[i].mark)) writes(" CHECKED");
      writes("></TD></TR>\r\n");
   }
   writes("</TABLE><P STYLE=\"margin-top: 0;\"><BR></P>\r\n");
   writes("<TABLE BORDER><CAPTION>Ident/IP filters</CAPTION>\r\n");
   writes("<TR><TH>Ident mask</TH><TH>IP mask</TH><TH>Flags</TH><TH></TH></TR>\r\n");
   for(i=0;i<entry.identipcount;i++)
   {
      writes("<TR><TD><INPUT TYPE=\"text\" NAME=\"ident\" SIZE=\"10\"");
      if(entry.identip[i].ident) writes(" VALUE=\"%s\"",entry.identip[i].ident);
      writes("></TD><TD><INPUT TYPE=\"text\" NAME=\"ip\" SIZE=\"20\"");
      if(entry.identip[i].ip) writes(" VALUE=\"%s\"",entry.identip[i].ip);
      writes("><TD>Control<INPUT TYPE=\"checkbox\" NAME=\"cflag\"");
      if(entry.identip[i].flags&ID_CONTROL) writes(" CHECKED");
      writes("> Data<INPUT TYPE=\"checkbox\" NAME=\"dflag\"");
      if(entry.identip[i].flags&ID_DATA) writes(" CHECKED");
      writes("> Ban<INPUT TYPE=\"checkbox\" NAME=\"bflag\"");
      if(entry.identip[i].flags&ID_BAN) writes(" CHECKED");
      writes("></TD><TD><INPUT TYPE=\"submit\" NAME=\"delidentip\" VALUE=\"Remove\"></TD></TR>\r\n",entry.uid,i);
   }
   writes("<TR><TD COLSPAN=3 ALIGN=CENTER><INPUT TYPE=\"submit\" NAME=\"addidentip\" VALUE=\"Add\"></TD></TR>\r\n");
   writes("</TABLE><P STYLE=\"margin-top: 0;\"><BR><HR><P STYLE=\"margin-top: 0\">\r\n");
   writes("<A HREF=\"/userconfig\">Back</A>\r\n");
   writes("<INPUT TYPE=\"submit\" VALUE=\"Save changes\">\r\n");
   writes("<INPUT TYPE=\"reset\" VALUE=\"Reset\">\r\n");
   if(strcmp(entry.name,"ftp"))
      writes("<INPUT TYPE=\"submit\" NAME=\"chpass\" VALUE=\"Change password\">\r\n");
   writes("<INPUT TYPE=\"submit\" NAME=\"deluser\" VALUE=\"Delete this user\"></P></FORM>\r\n");
   writes("</BODY></HTML>\r\n");
}

void usrlist(char *args)
{
   struct dftp_uent entry; int result;
   char *identip;

   if(!user.flags || !strchr(user.flags,'S'))
      cmd_err("403 Forbidden","Operator access required to access requested document",NULL);
   if(args)
   {
      usredit(args); return;
   }
   writes("HTTP/1.0 200 OK\r\n");
   writes(_headers_);
   writes(_htmltype_);
   writes("<HTML><HEAD><TITLE>Demi-FTPd User Configuration </TITLE></HEAD>\r\n");
   writes("<BODY STYLE=\"text-align: center; background-color: silver;\"><TABLE BORDER>\r\n");
   writes("<TR><TH>User name</TH>\r\n");
   writes("<TH>Flags</TH>\r\n");
   writes("<TH>Max connections</TH>\r\n");
   writes("<TH>User ID</TH>\r\n");
   writes("<TH>Group ID</TH>\r\n");
   writes("<TH>Ratio</TH>\r\n");
   writes("<TH>Uploaded bytes</TH>\r\n");
   writes("<TH>Downloaded bytes</TH>\r\n");
   writes("<TH>Credits</TH>\r\n");
   writes("<TH>Ident/IP filters</TH></TR>\r\n");
   if(httpopencfgfile(&passwdfile,USERFILE,O_RDWR))
   {
      flock(passwdfile,LOCK_EX); lseek(passwdfile,0,SEEK_SET);
      while((result = dftp_ugetent(passwdfile,&entry)))
         if(result < 0)
         {
            writes("<TR><TD ALIGN=\"center\"><A HREF=\"userconfig?user=%hu\">",entry.uid);
            writes("%s</A></TD>\r\n",entry.name);
            writes("<TD ALIGN=\"center\">");
            if(entry.flags) writes("%s",entry.flags);
            writes("</TD><TD ALIGN=\"center\">%hu</TD>",entry.maxconn);
            writes("<TD ALIGN=\"center\">%hu</TD>",entry.uid);
            writes("<TD ALIGN=\"center\">%hu</TD>",entry.gid);
            writes("<TD ALIGN=\"center\">1:%hu</TD>",entry.ratio);
            writes("<TD ALIGN=\"center\">%s</TD>",dftp_printbytes(entry.ulbytes));
            writes("<TD ALIGN=\"center\">%s</TD>",dftp_printbytes(entry.dlbytes));
            writes("<TD ALIGN=\"center\">%s</TD>",dftp_printbytes(entry.credits));
            identip = dftp_identip2text(entry.identip,entry.identipcount);
            if(identip) writes("<TD ALIGN=\"left\">%s</TD></TR>",identip);
            free(identip);
        }
      flock(passwdfile,LOCK_UN);
   }
   writes("</TABLE>\r\n"); 
   writes("<FORM METHOD=\"POST\" ACTION=\"userconfig\"><HR><P STYLE=\"margin-top: 0;\">\r\n");
   writes("<A HREF=\"/\">Back</A>\r\n");
   writes("<INPUT TYPE=\"hidden\" NAME=\"user\" VALUE=\"0\">");
   writes("<INPUT TYPE=\"submit\" VALUE=\"Add user\"></P></FORM>\r\n");
   writes("</BODY></HTML>\r\n");
}

/* Group information handling related routines */

unsigned int getnextid()
{
   unsigned int id = FIRSTGID, *ids = NULL; unsigned int idcount = 0; int i;
   struct dftp_gent entry;

   if(!httpopencfgfile(&groupfile,GROUPFILE,O_RDWR)) return id;
   flock(groupfile,LOCK_EX); lseek(groupfile,0,SEEK_SET);
   while(dftp_ggetent(groupfile,&entry))
   {
      ids = realloc(ids,++idcount*sizeof(unsigned int));
      ids[idcount-1] = entry.id;
      dftp_gfree(&entry);
   }
   flock(groupfile,LOCK_UN);
   for(i=0;i<idcount;i++)
      if(id == ids[i])
      {
         id++; i = -1;
      }
   return id;
}

void grplist(char *);

void getgroup(char *args, struct dftp_gent *entry)
{
   int result; char *token, *value, newmember[NAMELEN+1]; unsigned short gnum = 0;
   struct dftp_uent newmementry;
   
   args = parseargs(args);
   value = strsep(&args,"&"); if(!value) value = args;
   token = strsep(&value,"=");
   if(token && !strcmp(token,"group")) gnum = atoi(value);
   if(!args || !gnum)
   {
      if(!gnum)
      {
         entry->id = getnextid(); sprintf(entry->name,"group%hu",entry->id-FIRSTGID+1);
         if(!httpopensitefile(&groupsitefile,"group",O_RDWR))
            cmd_err("500 Internal server error","Cannot access in-site group database",NULL);
         flock(groupfile,LOCK_EX); lseek(groupfile,0,SEEK_END);
         dftp_gaddent(groupfile,entry); flock(groupsitefile,LOCK_EX);
         dftp_gupdatesitedb(groupfile,groupsitefile);
         flock(groupfile,LOCK_UN); flock(groupsitefile,LOCK_UN);
         return;
      }
      if(!httpopencfgfile(&groupfile,GROUPFILE,O_RDONLY))
         cmd_err("500 Internal server error","Cannot complete your request.",NULL);
      flock(groupfile,LOCK_EX); lseek(groupfile,0,SEEK_SET);
      result = dftp_ggetbyid(groupfile,gnum,entry);
      flock(groupfile,LOCK_UN);
      if(!result) cmd_err("404 Not found","Requested group not found",NULL);
      return;
   }
   entry->id = gnum;
   if(!httpopencfgfile(&groupfile,GROUPFILE,O_RDWR))
      cmd_err("500 Internal server error","Cannot access group database",NULL);
   if(!httpopensitefile(&groupsitefile,"group",O_RDWR))
      cmd_err("500 Internal server error","Cannot access in-site group database",NULL);
   while(args && *args)
   {
      value = strsep(&args,"&"); if(!value) value = args;
      if(strchr(value,'='))
      {
         token = strsep(&value,"=");
         if(!strcmp(token,"name")) strncpy(entry->name,value,NAMELEN);
         else if(!strcmp(token,"member"))
         {
            entry->members = realloc(entry->members,++entry->membercount*sizeof(char *));
            entry->members[entry->membercount-1] = malloc(strlen(value)+1);
            strcpy(entry->members[entry->membercount-1],value);
         }
         else if(!strcmp(token,"delmember"))
         {
            /* Workaround for realloc() bug found in glibc */
            if(--entry->membercount)
               entry->members = realloc(entry->members,entry->membercount*sizeof(char *));
            else
            {
               free(entry->members); entry->members = NULL;
            }
         }
         else if(!strcmp(token,"delgroup"))
         {
            flock(groupfile,LOCK_EX); result = dftp_gdelent(groupfile,gnum);
            if(!result) cmd_err("404 Not found","Target group not found",NULL);
            flock(groupsitefile,LOCK_EX); dftp_gupdatesitedb(groupfile,groupsitefile);
            flock(groupfile,LOCK_UN); flock(groupsitefile,LOCK_UN);
            grplist(NULL); exit(0);
         }
         else if(!strcmp(token,"newmember")) strcpy(newmember,value);
         else if(!strcmp(token,"addmember"))
         {
            if(!httpopencfgfile(&passwdfile,USERFILE,O_RDONLY))
               cmd_err("404 Not found","Cannot add member - cannot access user database",NULL);
            flock(passwdfile,LOCK_EX); lseek(passwdfile,0,SEEK_SET);
            result = dftp_ugetbyname(passwdfile,newmember,&newmementry); flock(groupfile,LOCK_UN);
            if(!result)
               cmd_err("404 Not found","Cannot add member - specified user does not exist",NULL);
            entry->members = realloc(entry->members,++entry->membercount*sizeof(char *));
            entry->members[entry->membercount-1] = malloc(strlen(newmember)+1);
            strcpy(entry->members[entry->membercount-1],newmember);
         }
      }
   }
   if(strcspn(entry->name,invalidchars) < strlen(entry->name))
      cmd_err("409 Conflict","Invalid character(s) in name",NULL);
   flock(groupfile,LOCK_EX); lseek(groupfile,0,SEEK_SET);
   dftp_gupdatedb(groupfile,entry); flock(groupsitefile,LOCK_EX);
   dftp_gupdatesitedb(groupfile,groupsitefile);
   flock(groupfile,LOCK_UN); flock(groupsitefile,LOCK_UN);
}

void grpedit(char *args)
{
   struct dftp_gent entry; int i, hasprimaryaffils = 0;

   memset(&entry,0,sizeof(entry));
   getgroup(args,&entry);
   writes("HTTP/1.0 200 OK\r\n");
   writes(_headers_);
   writes(_htmltype_);
   writes("<HTML><HEAD><TITLE>Demi-FTPd Group Configuration </TITLE></HEAD>\r\n");
   writes("<BODY STYLE=\"text-align: center; font-size: 18pt; background-color: silver;\">\r\n");
   writes("<FORM METHOD=\"POST\" ACTION=\"groupconfig\"><P>\r\n");
   writes("<INPUT TYPE=\"hidden\" NAME=\"group\" VALUE=\"%hu\">\r\n",entry.id);
   if(strcmp(entry.name,"ftp"))
      writes("Name <INPUT TYPE=\"text\" NAME=\"name\" MAXLENGTH=\"%i\" SIZE=\"%i\" VALUE=\"%s\">\r\n",NAMELEN,NAMELEN,entry.name);
   else writes("<B>Anonymous group</B><INPUT TYPE=\"hidden\" NAME=\"name\" VALUE=\"ftp\"> ");
   writes("<BR><SPAN STYLE=\"font-size: smaller;\">ID=%hu</SPAN><BR></P>\r\n",entry.id);
   writes("<TABLE BORDER><CAPTION>Members</CAPTION>\r\n");
   writes("<TR><TH>Name</TH><TH></TH></TR>\r\n");
   if(httpopencfgfile(&passwdfile,USERFILE,O_RDONLY))
   {
      struct dftp_uent uentry;

      flock(passwdfile,LOCK_EX); lseek(passwdfile,0,SEEK_SET);
      while((i=dftp_ugetent(passwdfile,&uentry)))
         if(i < 0 && uentry.gid == entry.id)
         {
            writes("<TR><TD>%s</TD></TR>\r\n",uentry.name);
            dftp_ufree(&uentry); hasprimaryaffils = 1;
         }
      flock(passwdfile,LOCK_UN); close(passwdfile);
   }
   for(i=0;i<entry.membercount;i++)
   {
      writes("<TR><TD><INPUT TYPE=\"hidden\" NAME=\"member\" VALUE=\"%s\">",entry.members[i]);
      writes("<A HREF=\"userconfig?user=%s\">%s</A></TD>\r\n",entry.members[i],entry.members[i]);
      writes("<TD><INPUT TYPE=\"submit\" NAME=\"delmember\" VALUE=\"Remove\"></TD></TR>\r\n",i);
   }
   writes("</TABLE><P>\r\n");
   writes("<INPUT TYPE=\"text\" NAME=\"newmember\" SIZE=\"%i\" MAXLENGTH=\"%i\"> ",NAMELEN,NAMELEN);
   writes("<INPUT TYPE=\"submit\" NAME=\"addmember\" VALUE=\"Add member\">");
   writes("<HR><P STYLE=\"margin-top: 0;\">\r\n");
   writes("<A HREF=\"/groupconfig\">Back</A>\r\n");
   writes("<INPUT TYPE=\"submit\" VALUE=\"Save changes\">\r\n");
   writes("<INPUT TYPE=\"reset\" VALUE=\"Reset\">\r\n");
   if(!hasprimaryaffils)
      writes("<INPUT TYPE=\"submit\" NAME=\"delgroup\" VALUE=\"Remove this group\">\r\n");
   writes("</P></FORM></BODY></HTML>\r\n");
}

void grplist(char *args)
{
   struct dftp_gent entry; int i, result;

   if(!user.flags || !strchr(user.flags,'S'))
      cmd_err("403 Forbidden","Operator access required to access requested document",NULL);
   if(args)
   {
      grpedit(args); return;
   }
   writes("HTTP/1.0 200 OK\r\n");
   writes(_headers_);
   writes(_htmltype_);
   writes("<HTML><HEAD><TITLE>Demi-FTPd Group Configuration </TITLE></HEAD>\r\n");
   writes("<BODY STYLE=\"text-align: center; background-color: silver;\">\r\n");
   writes("<FORM METHOD=\"POST\" ACTION=\"groupconfig\"><TABLE BORDER>\r\n");
   writes("<TR><TH><B>Group name</B></TH>\r\n");
   writes("<TH><B>ID</B></TH>\r\n");
   writes("<TH><B>Members</B></TH></TR>\r\n");
   if(httpopencfgfile(&groupfile,GROUPFILE,O_RDWR))
   {
      flock(groupfile,LOCK_EX); lseek(groupfile,0,SEEK_SET);
      while((result = dftp_ggetent(groupfile,&entry)))
         if(result < 0)
         {
            writes("<TR><TD ALIGN=\"center\"><A HREF=\"groupconfig?group=%hu\">",entry.id);
            writes("%s</A></TD>",entry.name);
            writes("<TD ALIGN=\"center\">%hu</TD><TD ALIGN=\"left\">",entry.id);
            for(i=0;i<entry.membercount;i++)
            {
               if(i) writes(",");
               writes("%s",entry.members[i]);
            }
        }
      flock(groupfile,LOCK_UN);
   }
   writes("</TABLE><P STYLE=\"margin-top: 0;\"><BR></P>\r\n");
   writes("<HR><P STYLE=\"margin-top: 0;\"><A HREF=\"/\">Back</A>\r\n");
   writes("<INPUT TYPE=\"hidden\" NAME=\"group\" VALUE=\"0\">");
   writes("<INPUT TYPE=\"submit\" VALUE=\"Add group\"></P></FORM></BODY></HTML>\r\n");
}

/* General settings handling related routines */

void updatesettings(char *args)
{
   char *token, *value; struct dftp_config cfg2;

   args = parseargs(args);
   memset(&cfg2,0,sizeof(struct dftp_config));
   cfg2.confdir = malloc(strlen(cfg.confdir)+1); strcpy(cfg2.confdir,cfg.confdir);
   while(args && *args)
   {
      value = strsep(&args,"&"); if(!value) value = args;
      if(strchr(value,'='))
      {
         token = strsep(&value,"=");
         if(!strcmp(token,"sitedir"))
         {
            cfg2.sitedir = malloc(strlen(value)+1); strcpy(cfg2.sitedir,value);
         }
         else if(!strcmp(token,"port")) cfg2.port = atoi(value);
         else if(!strcmp(token,"httpport")) cfg2.httpport = atoi(value);
         else if(!strcmp(token,"debuglevel")) cfg2.debuglevel = atoi(value);
         else if(!strcmp(token,"mod_proctitle")) cfg2.mod_proctitle = 1;
         else if(!strcmp(token,"resolvehosts")) cfg2.resolvehosts = 1;
         else if(!strcmp(token,"realusersupport")) cfg2.realusersupport = 1;
         else if(!strcmp(token,"shadow")) cfg2.shadow = 1;
         else if(!strcmp(token,"maxviewsize")) cfg2.maxviewsize = atoi(value);
         else if(!strcmp(token,"maxidletime")) cfg2.maxidletime = atoi(value);
         else if(!strcmp(token,"loginidletime")) cfg2.loginidletime = atoi(value);
         else if(!strcmp(token,"stupdatetime")) cfg2.stupdatetime = atoi(value);
         else if(!strcmp(token,"maxloginfails")) cfg2.maxloginfails = atoi(value);
         else if(!strcmp(token,"cmode")) sscanf(value,"%o",&cfg2.cmode);
         else if(!strcmp(token,"welcomefile"))
         {
            cfg2.welcomefile = malloc(strlen(value)+1); strcpy(cfg2.welcomefile,value);
         }
         else if(!strcmp(token,"goodbyemsg"))
         {
            cfg2.goodbyemsg = malloc(strlen(value)+1); strcpy(cfg2.goodbyemsg,value);
         }
         else if(!strcmp(token,"welcomemsg"))
         {
            cfg2.welcomemsg = malloc(strlen(value)+1); strcpy(cfg2.welcomemsg,value);
         }
         else if(!strcmp(token,"status"))
         {
            cfg2.statusline = malloc(strlen(value)+1); strcpy(cfg2.statusline,value);
         }
         else if(!strncmp(token,"logmask-",8))
         {
            cfg2.logs = realloc(cfg2.logs,++cfg2.logcount*sizeof(struct dftp_log));
            cfg2.logs[cfg2.logcount-1].pattern = malloc(strlen(value)+1);
            strcpy(cfg2.logs[cfg2.logcount-1].pattern,value);
         }
         else if(!strncmp(token,"logfile-",8) && value)
         {
            cfg2.logs[cfg2.logcount-1].path = malloc(strlen(value)+1);
            strcpy(cfg2.logs[cfg2.logcount-1].path,value);
         }
         else if(!strncmp(token,"ulcmask-",8))
         {
            cfg2.ulchecker = realloc(cfg2.ulchecker,++cfg2.uccount*sizeof(struct dftp_ulchecker));
            cfg2.ulchecker[cfg2.uccount-1].mask = malloc(strlen(value)+1);
            strcpy(cfg2.ulchecker[cfg2.uccount-1].mask,value);
         }
         else if(!strncmp(token,"ulchand-",8))
         {
            cfg2.ulchecker[cfg2.uccount-1].handler = malloc(strlen(value)+1);
            strcpy(cfg2.ulchecker[cfg2.uccount-1].handler,value);
         }
         else if(!strncmp(token,"ducmask-",8))
         {
            cfg2.dupechecker = realloc(cfg2.dupechecker,++cfg2.dccount*sizeof(char *));
            cfg2.dupechecker[cfg2.dccount-1] = malloc(strlen(value)+1);
            strcpy(cfg2.dupechecker[cfg2.dccount-1],value);
         }
         else if(!strcmp(token,"dellog"))
            cfg2.logs = realloc(cfg2.logs,--cfg2.logcount*sizeof(struct dftp_log));
         else if(!strcmp(token,"addlog"))
         {
            cfg2.logs = realloc(cfg2.logs,++cfg2.logcount*sizeof(struct dftp_log));
            cfg2.logs[cfg2.logcount-1].pattern = malloc(2);
            cfg2.logs[cfg2.logcount-1].path = malloc(strlen(cfg2.confdir)+11);
            sprintf(cfg2.logs[cfg2.logcount-1].pattern,"*");
            sprintf(cfg2.logs[cfg2.logcount-1].path,"%s/dftpd.log",cfg2.confdir);
         }
         else if(!strcmp(token,"deluc"))
            cfg2.ulchecker = realloc(cfg2.ulchecker,--cfg2.uccount*sizeof(struct dftp_ulchecker));
         else if(!strcmp(token,"adduc"))
         {
            cfg2.ulchecker = realloc(cfg2.ulchecker,++cfg2.uccount*sizeof(struct dftp_ulchecker));
            cfg2.ulchecker[cfg2.uccount-1].mask = malloc(2);
            cfg2.ulchecker[cfg2.uccount-1].handler = NULL;
            sprintf(cfg2.ulchecker[cfg2.uccount-1].mask,"*");
         }
         else if(!strcmp(token,"deldc"))
            cfg2.dupechecker = realloc(cfg2.dupechecker,--cfg2.dccount*sizeof(char *));
         else if(!strcmp(token,"adddc"))
         {
            cfg2.dupechecker = realloc(cfg2.dupechecker,++cfg2.dccount*sizeof(char *));
            cfg2.dupechecker[cfg2.dccount-1] = malloc(2);
            sprintf(cfg2.dupechecker[cfg2.dccount-1],"*");
         }
      }
   }
   dftp_freeconfig(&cfg); cfg = cfg2; if(!inetd) kill(getppid(),SIGHUP);
}

void settings(char *args)
{
   int i, cfgfile = -1, tempfd = -1;

   if(!user.flags || !strchr(user.flags,'S'))
      cmd_err("403 Forbidden","Operator access required to access requested document",NULL);
   if(args)
   {
      updatesettings(args);
      if(!httpopencfgfile(&cfgfile,CONFIGFILE,O_WRONLY|O_TRUNC))
         cmd_err("500 Internal server error","Cannot complete your request",NULL);
      flock(cfgfile,LOCK_EX); tempfd = sockfd; sockfd = cfgfile;
      writes("sitedir=%s\n",cfg.sitedir);
      writes("port=%hu\n",cfg.port);
      writes("httpport=%hu\n",cfg.httpport);
      writes("debuglevel=%hu\n",cfg.debuglevel);
      writes("createmode=%o\n",cfg.cmode);
      writes("maxloginfails=%hu\n",cfg.maxloginfails);
      writes("maxidletime=%hu\n",cfg.maxidletime);
      writes("loginidletime=%hu\n",cfg.loginidletime);
      writes("stupdatetime=%u\n",cfg.stupdatetime);
      writes("maxviewsize=%u\n",cfg.maxviewsize);
      for(i=0;i<cfg.logcount;i++)
      {
         writes("log %s=",cfg.logs[i].pattern);
         if(cfg.logs[i].path) writes("%s",cfg.logs[i].path);
         writes("\n");
      }
      for(i=0;i<cfg.uccount;i++)
      {
         writes("ulcheck %s=",cfg.ulchecker[i].mask);
         if(cfg.ulchecker[i].handler) writes("%s",cfg.ulchecker[i].handler);
         writes("\n");
      }
      for(i=0;i<cfg.dccount;i++)
         writes("dupecheck=%s\n",cfg.dupechecker[i]);
      if(cfg.welcomefile) writes("welcomefile=%s\n",cfg.welcomefile);
      if(cfg.welcomemsg) writes("welcomemsg=%s\n",cfg.welcomemsg);
      if(cfg.goodbyemsg) writes("goodbyemsg=%s\n",cfg.goodbyemsg);
      if(cfg.statusline) writes("statusline=%s\n",cfg.statusline);
      writes("proctitle=%i\n",cfg.mod_proctitle);
      writes("resolvehosts=%i\n",cfg.resolvehosts);
      writes("realusersupport=%i\n",cfg.realusersupport);
      writes("shadow=%i\n",cfg.shadow);
      flock(cfgfile,LOCK_UN); sockfd = tempfd;
   }
   writes("HTTP/1.0 200 OK\r\n");
   writes(_headers_);
   writes(_htmltype_);
   writes("<HTML><HEAD><TITLE>Demi-FTPd configuration</TITLE></HEAD>\r\n");
   writes("<BODY STYLE=\"text-align: center; font-size: 18pt; background-color: silver;\">\r\n");
   writes("<FORM METHOD=\"POST\" ACTION=\"settings\" STYLE=\"text-align: center;\">\r\n");
   writes("<H1>General server settings</H1><P>\r\n");
   writes("Site directory <INPUT TYPE=\"text\" NAME=\"sitedir\" SIZE=20 VALUE=\"%s\"><BR>\r\n",cfg.sitedir);
   writes("FTP server port <INPUT TYPE=\"text\" NAME=\"port\" SIZE=5 MAXLENGTH=5 VALUE=\"%hu\">\r\n",cfg.port);
   writes("DWAT port <INPUT TYPE=\"text\" NAME=\"httpport\" SIZE=5 MAXLENGTH=5 VALUE=\"%hu\"><BR>\r\n",cfg.httpport);
   writes("Debug level <INPUT TYPE=\"text\" NAME=\"debuglevel\" SIZE=1 MAXLENGTH=1 VALUE=\"%hu\">\r\n",cfg.debuglevel);
   writes("Create mode <INPUT TYPE=\"text\" NAME=\"cmode\" SIZE=3 MAXLENGTH=3 VALUE=\"%o\"><BR>\r\n",cfg.cmode);

   writes("<HR><H1>Switches</H1><P>\r\n");
   writes("Replace process title <INPUT TYPE=\"checkbox\" NAME=\"mod_proctitle\"");
   if(cfg.mod_proctitle) writes(" CHECKED");
   writes(">\r\n<BR>Host name resolving <INPUT TYPE=\"checkbox\" NAME=\"resolvehosts\"");
   if(cfg.resolvehosts) writes(" CHECKED");
   writes(">\r\n<BR>Allow real users <INPUT TYPE=\"checkbox\" NAME=\"realusersupport\"");
   if(cfg.realusersupport) writes(" CHECKED");
   writes(">\r\n<BR>Use shadow passwords <INPUT TYPE=\"checkbox\" NAME=\"shadow\"");
   if(cfg.shadow) writes(" CHECKED");

   writes(">\r\n<BR></P><HR><H1>Logging</H1>\r\n");
   writes("<TABLE BORDER>\r\n");
   writes("<TR><TH>Event mask</TH><TH>File</TH><TH></TH></TR>\r\n");
   for(i=0;i<cfg.logcount;i++)
   {
      writes("<TR><TD><INPUT TYPE=\"text\" NAME=\"logmask-%i\" VALUE=\"%s\"></TD>\r\n",i,cfg.logs[i].pattern);
      writes("<TD><INPUT TYPE=\"text\" NAME=\"logfile-%i\"",i);
      if(cfg.logs[i].path) writes(" VALUE=\"%s\"",cfg.logs[i].path);
      writes("></TD>\r\n",i,cfg.logs[i].path);
      writes("<TD><INPUT TYPE=\"submit\" NAME=\"dellog\" VALUE=\"Remove\"></TD></TR>\r\n");
   }
   writes("<TR><TD COLSPAN=\"2\" ALIGN=\"center\"><INPUT TYPE=\"submit\" NAME=\"addlog\" VALUE=\"Add\"></TD></TR>\r\n");
   writes("</TABLE><P STYLE=\"margin-top: 0;\"><BR></P><HR>\r\n");

   writes("<H1>Limits</H1><P>\r\n");
   writes("Max login failures <INPUT TYPE=\"text\" NAME=\"maxloginfails\" SIZE=2 MAXLENGTH=2 VALUE=\"%hu\"><BR>\r\n",cfg.maxloginfails);
   writes("Maximum idle time(in seconds)<BR>\r\n");
   writes("during login time <INPUT TYPE=\"text\" NAME=\"loginidletime\" SIZE=4 MAXLENGTH=4 VALUE=\"%hu\"> \r\n",cfg.loginidletime);
   writes("after login <INPUT TYPE=\"text\" NAME=\"maxidletime\" SIZE=4 MAXLENGTH=4 VALUE=\"%hu\"><BR>\r\n",cfg.maxidletime);
   writes("Maximum file size(in bytes) viewable by SITE VIEW ");
   writes("<INPUT TYPE=\"text\" NAME=\"maxviewsize\" SIZE=9 MAXLENGTH=9 VALUE=\"%u\">\r\n",cfg.maxviewsize);
   writes("<BR>Transfer status update interval(in seconds) <INPUT TYPE=\"text\" SIZE=2 MAXLENGTH=5 NAME=\"stupdatetime\" VALUE=\"%u\">\r\n",cfg.stupdatetime);

   writes("<BR><HR><H1>Server messages</H1>\r\n");
   writes("<TABLE BORDER=\"0\">\r\n");
   writes("<TR><TD>File shown after successful login</TD><TD><INPUT TYPE=\"text\" NAME=\"welcomefile\" SIZE=30");
   if(cfg.welcomefile) writes(" VALUE=\"%s\"",cfg.welcomefile);
   writes("></TD></TR>\r\n<TR><TD ALIGN=RIGHT>Welcome message</TD>");
   writes("<TD><INPUT TYPE=\"text\" NAME=\"welcomemsg\" SIZE=30");
   if(cfg.welcomemsg) writes(" VALUE=\"%s\"",cfg.welcomemsg);
   writes("></TD></TR>\r\n");
   writes("<TR><TD ALIGN=RIGHT>Goodbye message</TD>");
   writes("<TD><INPUT TYPE=\"text\" NAME=\"goodbyemsg\" SIZE=30");
   if(cfg.goodbyemsg) writes(" VALUE=\"%s\"",cfg.goodbyemsg);
   writes("></TD></TR>\r\n");
   writes("<TR><TD ALIGN=RIGHT>Status response</TD>");
   writes("<TD><INPUT TYPE=\"text\" NAME=\"status\" SIZE=30");
   if(cfg.statusline) writes(" VALUE=\"%s\"",cfg.statusline);
   writes("></TD></TR></TABLE>\r\n");

   writes("<HR><H1>Upload & Duplicate checker settings</H1>\r\n");
   writes("<TABLE BORDER><TR><TH COLSPAN=3>Upload checker</TH><TH COLSPAN=2>Duplicate checker</TH></TR>\r\n");
   writes("<TR><TH>Filemask</TH><TH>Handler</TH><TH></TH><TH>Filemask</TH></TR><TR><TD></TD></TR>\r\n");
   for(i=0;i<cfg.uccount || i<cfg.dccount;i++)
   {
      if(i<cfg.uccount)
      {
         writes("<TR><TD><INPUT TYPE=\"text\" NAME=\"ulcmask-%i\" VALUE=\"%s\"></TD>",i,cfg.ulchecker[i].mask);
         writes("<TD><INPUT TYPE=\"text\" NAME=\"ulchand-%i\"",i);
         if(cfg.ulchecker[i].handler) writes(" VALUE=\"%s\"",cfg.ulchecker[i].handler);
         writes("></TD>",i,cfg.ulchecker[i].handler);
         writes("<TD><INPUT TYPE=\"submit\" NAME=\"deluc\" VALUE=\"Remove\"></TD>");
      }
      else writes("<TR><TD></TD><TD></TD><TD></TD>");
      if(i<cfg.dccount)
      {
         writes("<TD><INPUT TYPE=\"text\" NAME=\"ducmask-%i\" VALUE=\"%s\"></TD>",i,cfg.dupechecker[i]);
         writes("<TD><INPUT TYPE=\"submit\" NAME=\"deldc\" VALUE=\"Remove\"></TD></TR>\r\n");
      }
      else writes("<TD></TD><TD></TD></TR>\r\n");
   }
   writes("<TR><TD COLSPAN=2 ALIGN=CENTER><INPUT TYPE=\"submit\" NAME=\"adduc\" VALUE=\"Add\"></TD><TD></TD>\r\n");
   writes("<TD COLSPAN=1 ALIGN=CENTER><INPUT TYPE=\"submit\" NAME=\"adddc\" VALUE=\"Add\"></TD></TR>\r\n");
   writes("</TABLE><P STYLE=\"margin-top: 0;\"><BR><P><HR><P STYLE=\"margin-top: 0;\">\r\n");
   writes("<A HREF=\"/\">Back</A> \r\n");
   writes("<INPUT TYPE=\"submit\" VALUE=\"Save changes\">\r\n");
   writes("<INPUT TYPE=\"reset\" VALUE=\"Reset\"></P></FORM>\r\n");
   writes("</BODY></HTML>\r\n");
}

/* Main screen */

void indexpage()
{
   writes("HTTP/1.0 200 OK\r\n");
   writes(_headers_);
   writes(_htmltype_);
   writes("<HTML><HEAD>\r\n");
   writes("<TITLE>Demi-FTPd configuration</TITLE>\r\n");
   writes("</HEAD><BODY STYLE=\"text-align: center; font-size: 20pt; background-color: silver;\">\r\n");
   writes("<H1 STYLE=\"text-decoration: underline;\">Dftpd v%s configuration</H1><P>\r\n",VER);
   writes("<A HREF=\"monitor\">Monitor usage</A><BR>\r\n");
   if(user.flags && strchr(user.flags,'S')) writes("<A HREF=\"userconfig\">");
   writes("Configure users");
   if(user.flags && strchr(user.flags,'S')) writes("</A>");
   writes("<BR>\r\n");

   if(user.flags && strchr(user.flags,'S')) writes("<A HREF=\"groupconfig\">");
   writes("Configure groups");
   if(user.flags && strchr(user.flags,'S')) writes("</A>");
   writes("<BR>\r\n");

   if(user.flags && strchr(user.flags,'S')) writes("<A HREF=\"settings\">");
   writes("Change server settings");
   if(user.flags && strchr(user.flags,'S')) writes("</A>");
   writes("<BR>\r\n");

   writes("</P><HR><P STYLE=\"font-size: 14pt; margin-top: 0;\">\r\n");
   writes("Demi-FTPd Copyright 1999 Demi-god<BR>\r\n");
   writes("<SPAN STYLE=\"font-style: italic;\">Warning: this program transmits information insecurely.<BR>");
   writes("DO NOT use it over insecure networks!</SPAN></P></BODY></HTML>");
}

void base64decode(char *str)
{
   static const char *b64 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
   int i, j, byteorder = 0, numorder = 3; unsigned long number = 0;

   for(i=0;i<strlen(str);i++)
   {
      if(str[i] == '=')
      {
         if(str[i+1] == '=') i = 2; else i = 1;
         for(j=0;j<3-i;j++) str[byteorder+j] = ((char *)&number)[2-j];
         byteorder += 3-i; break;
      }
      else number |= (unsigned long)(strchr(b64,str[i])-b64) << 6*numorder;
      if(!numorder)
      {
         for(j=0;j<3;j++) str[byteorder+j] = ((char *)&number)[2-j];
         number = 0; numorder = 3; byteorder += 3;
      }
      else numorder--;
   }
   str[byteorder] = 0;
}

int authorize()
{
   int result;

   if(!password || !strlen(user.name) || !strcmp(user.name,"ftp")) return 0;
   if(!httpopencfgfile(&passwdfile,USERFILE,O_RDWR)) return 0;
   flock(passwdfile,LOCK_EX); lseek(passwdfile,0,SEEK_SET);
   result = dftp_ugetbyname(passwdfile,user.name,&user);
   flock(passwdfile,LOCK_UN);
   if(!result) return 0;
   if(checkidentip(&cli_addr,ID_CONTROL))
      cmd_err("403 Forbidden","IP check failed - cannot access the requested document",NULL);
   if(!user.maxconn)
   {
      dftp_ufree(&user); return 0;
   } 
   if(strcmp(user.passwd,crypt(password,user.passwd)))
   {
      ftplog("AUTH","HTTP: Access denied for user %s: wrong password",user.name);
      dftp_ufree(&user); return 0;
   } 
   if(cfg.debuglevel >= 2) ftplog("AUTH","HTTP: DEBUG: Authorization: %s",user.name);
   return 1;
}

void getmeta(char *opt, char *param)
{
   char *lineptr = param;

   if(!strcmp(opt,"Authorization"))
   {
      char *uname = NULL;

      lineptr = param+6; base64decode(param+6);
      uname = strsep(&lineptr,":");
      strncpy(user.name,uname,NAMELEN); user.name[NAMELEN] = 0;
      password = malloc(strlen(lineptr)+1); strcpy(password,lineptr);
   }
}

void httpserve()
{
   char cmdline[MAXLINE], *lineptr = cmdline, *cmd = NULL, *param = NULL, junkline[MAXLINE];

   if(cfg.mod_proctitle) inst_str("Demi-FTPd - DWAT client");
   if(cfg.debuglevel >= 1) ftplog("CONN","HTTP: DEBUG: Incoming connection from %s",inet_ntoa(cli_addr.sin_addr));
   signal(SIGSEGV,httpclienterr);
   readline(sockfd,cmdline);
   do
   {
      lineptr = junkline;
      readline(sockfd,junkline);
      cmd = strsep(&lineptr,":"); param = lineptr+1;
      getmeta(cmd,param);
   } while(strlen(junkline));
   lineptr = cmdline;
   cmd = strsep(&lineptr," "); param = strsep(&lineptr," HTTP");
   ftplog("CMD","HTTP: %s %s",cmd,param);
   if(!authorize()) cmd_err("401 Unauthorized",
    "Authorization is required to access the requested document",
    "WWW-Authenticate: Basic realm=\"DFTPd configuration/monitoring\"");
   loadplugins();
   if(!strcmp(cmd,"GET"))
   {
      cmd = strchr(param,'?');
      if(cmd)
      {
         *cmd = 0; cmd++;
      }
      if(!strcmp(param,"/")) indexpage();
      else if(!strcmp(param,"/monitor")) monitor(cmd);
      else if(!strcmp(param,"/userconfig")) usrlist(cmd);
      else if(!strcmp(param,"/groupconfig")) grplist(cmd);
      else if(!strcmp(param,"/settings")) settings(cmd);
      else cmd_err("404 Not found","The document you requested was not found.",NULL);
   }
   else if(!strcmp(cmd,"POST"))
   {
      readline(sockfd,junkline);
      if(!strcmp(param,"/monitor")) monitor(junkline);
      else if(!strcmp(param,"/userconfig")) usrlist(junkline);
      else if(!strcmp(param,"/groupconfig")) grplist(junkline);
      else if(!strcmp(param,"/settings")) settings(junkline);
      else cmd_err("404 Not found","The document you requested was not found.",NULL);
   }
   else if(!strcmp(cmd,"HEAD"))
   {
      writes("HTTP/1.0 200 OK"); writes(_headers_);
   }
   else cmd_err("501 Unsupported","Command is not supported.",NULL);
   exit(0);
}

int main(int argc, char *argv[])
{
   int childpid, clilen, option, quiet = 0;

   dftp_readconfig(&cfg);

   while( (option = getopt(argc,argv,"qip:d:")) != EOF)
   {
      if(option == 'q') quiet = -1;
      if(option == 'i') inetd = -1;
      if(option == 'd') cfg.debuglevel = atoi(optarg);
      if(option == 'p') cfg.httpport = atoi(optarg);
      if(option == '?') return 1;
   }

   if(!quiet && !inetd) printf("Demi-FTPd Web Administration Tool server for dftpd v%s\n",VER);
   openlogs(quiet|inetd);

   Argv = argv; Argc = argc; memset(&user,0,sizeof(user));
   if(inetd)
   {
      quiet = sizeof(cli_addr);
      getpeername(0,&cli_addr,&quiet);
      sockfd = 0; httpserve(); exit(0);
   }
   if(cfg.mod_proctitle) inst_str("Demi-FTPd - DWAT server");
   memset(&serv_addr,0,sizeof(serv_addr));
   serv_addr.sin_family = AF_INET;
   serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
   serv_addr.sin_port = htons(cfg.httpport);
   if( (serv_sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
   {
      fprintf(stderr,"Can't open stream socket\n");
      return 1;
   }
   if(!quiet) printf("Binding to port %hu...",cfg.httpport); fflush(stdout);
   if(bind(serv_sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
   {
      fprintf(stderr,"Can't bind local address(%s)\n",strerror(errno));
      return 1;
   }
   if(!quiet) printf("done\n");
   signal(SIGPIPE,httplostconn); signal(SIGSEGV,httperr);
   signal(SIGTERM,httpquit); signal(SIGHUP,reloadconf);
   if(signal(SIGCLD,SIG_IGN) == SIG_ERR)
   {
      fprintf(stderr,"Fatal: could not ignore SIGCLD signal\n");
      close(serv_sockfd); return 8;
   }
   listen(serv_sockfd,BACKLOG); daemonize();
   ftplog("START","HTTP: Server started on port %hu",cfg.port+1);
   for(;;)
   {
      clilen = sizeof(cli_addr);
      sockfd = accept(serv_sockfd, (struct sockaddr *) &cli_addr, &clilen);
      if(sockfd >= 0)
      {
         childpid = fork();
         if(childpid == 0)
         {
            close(serv_sockfd);
            httpserve(); exit(0);
         }
         close(sockfd);
      }
   }
}


