/*  Copyright (C) 1999  Demi-god / Radical Corruption
    Demi-FTPd monitoring utility

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>
#include <signal.h>
#include <limits.h>
#include <ctype.h>
#include <curses.h>
#include "dftpd.h"

#define USERSLEN 16
#define PIDLEN 6
#define ONLINELEN 7
#define ORIGINLEN 25
#define ENTRIES LINES-4

int dftp_utmpgetent(int, struct dftp_utmpent *);
void handlealarm(int);

int fd = -1; struct dftp_utmpent **list = NULL;
unsigned int num = 0, xoffset = 0, yoffset = 0, wwidth, fieldlen[6];
WINDOW *pad = NULL;
const char titletext[] = "Demi-FTPd WHO utility v1.2";
char *fieldtext[] = { "User",
                            "Pid",
                            "Online",
                            "Activity",
                            "Host name",
                            "Data host name" };

char *printtime(time_t time1)
{
   time_t time2; int tot_secs,secs = 0,mins = 0,hrs = 0,days = 0;
   static char value[25];

   time(&time2); tot_secs = (int)difftime(time2,time1);
   days = tot_secs/84600; tot_secs -= days*86400;
   hrs = tot_secs/3600; tot_secs -= hrs*3600;
   mins = tot_secs/60; tot_secs -= mins*60;
   secs = tot_secs; memset(value,0,25);
   if(days) sprintf(value,"%id ",days);
   if(hrs) sprintf(value+strlen(value),"%ih ",hrs);
   if(mins) sprintf(value+strlen(value),"%im ",mins);
   sprintf(value+strlen(value),"%is",secs);
   return value;
}

void draw_screen()
{
   int i,pos; char tempstr[30];

   if(pad)
   {
      delwin(pad); pad = NULL;
   }
   for(i=0;i<6;i++) fieldlen[i] = strlen(fieldtext[i]);

   /* Determine field lengths */

   memset(tempstr,0,25);

   for(i=0;i<num;i++)
   {
      if(strlen(list[i]->name) > fieldlen[0]) fieldlen[0] =
       strlen(list[i]->name);
      sprintf(tempstr,"%i",list[i]->pid);
      if(strlen(tempstr) > fieldlen[1]) fieldlen[1] = 
       strlen(tempstr);
      strcpy(tempstr,printtime(list[i]->aika));
      if(strlen(tempstr) > fieldlen[2]) fieldlen[2] = 
       strlen(tempstr);
      if(!strlen(list[i]->activity))
      {
         sprintf(tempstr,"Idle(%s)",printtime(list[i]->idleaika));
         if(strlen(tempstr) > fieldlen[3]) fieldlen[3] = strlen(tempstr);
      }
      else if(strlen(list[i]->activity) > fieldlen[3]) fieldlen[3] = 
       strlen(list[i]->activity);
      if(strlen(list[i]->controlhostname) > fieldlen[4]) fieldlen[4] =
       strlen(list[i]->controlhostname);
      if(strlen(list[i]->datahostname) > fieldlen[5]) fieldlen[5] =
       strlen(list[i]->datahostname);
   }
   for(wwidth=5,i=0;i<6;i++) wwidth += fieldlen[i];
   if(wwidth < COLS-2) wwidth = COLS-2;
   pad = newpad(LINES-2,wwidth); wattrset(pad,COLOR_PAIR(1));
   for(i=0;i<(LINES-2)*wwidth;i++) waddch(pad,' ');
   move(0,0);
   if(has_colors())
   {
      wmove(pad,1,0); whline(pad, ACS_HLINE,wwidth);
      wmove(pad,0,0); wprintw(pad, fieldtext[0]);
      for(i=1,pos=fieldlen[0];i<6;i++)
      {
         wmove(pad,0,pos+1); wprintw(pad,fieldtext[i]); 
         wmove(pad,1,pos); wvline(pad,ACS_VLINE,LINES-3);
         waddch(pad,ACS_TTEE); pos += fieldlen[i]+1;
      }
   }
   else
   {
      wmove(pad,1,0); whline(pad, '-',wwidth);
      wmove(pad,0,0); wprintw(pad, fieldtext[0]);
      for(i=1,pos=fieldlen[0];i<6;i++)
      {
         wmove(pad,0,pos+1); wprintw(pad,fieldtext[i]); 
         wmove(pad,1,pos); wvline(pad,'|',LINES-3);
         waddch(pad,'+'); pos += fieldlen[i]+1;
      }
   }
}

void visual_setup()
{
   initscr(); cbreak(); noecho(); keypad(stdscr,TRUE);
   curs_set(0);
   if(has_colors())
   {
      start_color();
      init_pair(1,COLOR_WHITE,COLOR_BLUE);
      attrset(COLOR_PAIR(1));
      border(0,0,0,0,0,0,0,0);
      mvaddch(0,COLS/2-strlen(titletext)/2-1,ACS_RTEE);
      addstr(titletext); addch(ACS_LTEE);
   }
   else
   {
      border('|','|','-','-','+','+','+','+');
      mvaddch(0,COLS/2-strlen(titletext)/2-1,'|');
      addstr(titletext); addch('|');
   }
   refresh();
}

void free_list()
{
   while(num) free(list[--num]);
   free(list); list = NULL; num = 0;
}

void update_entries()
{
   int i,j,pos;

   draw_screen();
   if(xoffset+COLS-2 > wwidth) xoffset = wwidth-COLS+2;
   for(i=0,j=yoffset;i<ENTRIES && j<num;i++,j++)
   {
      pos = 0;
      mvwaddnstr(pad,2+i,pos,list[j]->name,fieldlen[0]); pos += fieldlen[0]+1;
      mvwprintw(pad,2+i,pos,"%hu",list[j]->pid); pos += fieldlen[1]+1;
      mvwprintw(pad,2+i,pos,"%s",printtime(list[j]->aika));
      pos += fieldlen[2]+1;
      if(!strlen(list[j]->activity))
         mvwprintw(pad,2+i,pos,"Idle(%s)",printtime(list[j]->idleaika));
      else mvwaddnstr(pad,2+i,pos,list[j]->activity,fieldlen[3]);
      pos += fieldlen[3]+1;
      mvwaddstr(pad,2+i,pos,list[j]->controlhostname);
      pos += fieldlen[4]+1;
      mvwaddnstr(pad,2+i,pos,list[j]->datahostname,fieldlen[5]);
   }
   prefresh(pad,0,xoffset,1,1,LINES-2,COLS-2);
}

void install_timer()
{
   signal(SIGALRM,handlealarm);
}

void handlealarm(int signal)
{
   struct dftp_utmpent entry; struct stat sbuf; int i,j;
   time_t changetime = 0; off_t koko = 0;

   fstat(fd,&sbuf);
   if(sbuf.st_ctime != changetime || sbuf.st_size != koko)
   {
      free_list();
      flock(fd,LOCK_EX);
      lseek(fd,0,SEEK_SET);
      while( (i=dftp_utmpgetent(fd,&entry)) )
         if(i < 0)
         {
            for(j=strlen(entry.name)-1;isspace(entry.name[j]) && j>=0;j--)
               entry.name[j] = 0;
            for(j=strlen(entry.activity)-1;isspace(entry.activity[j]) && j>=0;j--)
               entry.activity[j] = 0;
            for(j=strlen(entry.controlhostname)-1;isspace(entry.controlhostname[j]) && j>=0;j--)
               entry.controlhostname[j] = 0;
            for(j=strlen(entry.datahostname)-1;isspace(entry.datahostname[j]) && j>=0;j--)
               entry.datahostname[j] = 0;
            list = realloc(list,(num+1)*sizeof(void *));
            list[num] = malloc(sizeof(struct dftp_utmpent));
            memcpy(list[num++],&entry,sizeof(struct dftp_utmpent));
         }
      flock(fd,LOCK_UN);
      changetime = sbuf.st_ctime; koko = sbuf.st_size;
   }
   update_entries();
   install_timer();
}

void normal_dwho()
{
   int i,j = 0; struct dftp_utmpent entry;

   while( (i=dftp_utmpgetent(fd,&entry)) )
   {
      if(i < 0)
      {
         if(j)
         {
            printf("------------------------------------------------------------");
            printf("--------------------\n");
         }
         printf("User: ");
         if(!strcmp(entry.name,"ftp")) printf("<anonymous>      ");
         else if(strlen(entry.name) > 0) printf("%-16s ",entry.name);
         else printf("<null>          "); printf("Pid: %-5i ",entry.pid);
         printf("Online: %s\n",printtime(entry.aika));
         if(entry.activity[0] != 32) printf("Activity: %s\n",entry.activity);
         else printf("Activity: Idle(%s)\n",printtime(entry.idleaika));
         if(strlen(entry.controlhostname) > 0) printf("Control host name: %s\n",entry.controlhostname);
         else printf("Control host name: <NULL>\n");
         if(!isspace(entry.datahostname[0])) printf("Data host name: %s\n",entry.datahostname);
         j = 1;
      }
   }
   close(fd);
   exit(0);
}

void usage()
{
   printf("%s\n",titletext);
   printf("Usage: dwho [-c]\n");
   exit(0);
}

int main(int argc,char *argv[])
{
   int i; char utmpfilenimi[PATH_MAX];
   struct itimerval timer;

   if(argc > 1 && (!strcmp(argv[1],"--help") || !strcmp(argv[1],"-h"))) usage();
   if(getenv(DFTP_ENV)) sprintf(utmpfilenimi,"%s/%s",getenv(DFTP_ENV),UTMPFILE);
   else sprintf(utmpfilenimi,"%s/%s",conf_defaultpath,UTMPFILE);
   fd = open(utmpfilenimi,O_RDONLY);
   if(fd < 0)
   {
      fprintf(stderr,"Error opening %s for read\n",utmpfilenimi);
      return 1;
   }
   tzset();
   if(argc > 1 && !strcmp(argv[1],"-c")) normal_dwho();
   visual_setup();
   handlealarm(SIGALRM);
   timer.it_value.tv_sec = 1;
   timer.it_value.tv_usec = 0;
   timer.it_interval.tv_sec = 1;
   timer.it_interval.tv_usec = 0;

   setitimer(ITIMER_REAL,&timer,NULL);
   for(i=0;i != 'q';)
   {
      i = getch();
      if(i == KEY_UP && yoffset) { yoffset--; update_entries(); }
      else if(i == KEY_DOWN && yoffset+ENTRIES < num)
      {
         yoffset++; update_entries();
      }
      else if(i == KEY_LEFT && xoffset) { xoffset--; update_entries(); }
      else if(i == KEY_RIGHT && xoffset+COLS-2 < wwidth)
      {
         xoffset++; update_entries();
      }
   }
   clear(); refresh(); endwin(); close(fd); free_list();
   return 0;
}
