#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/file.h>
#include "dftpd.h"

//void ftplog(const char *, const char *, ...);
char *dftp_readln(const int fd);

static const char dftp_defaultflags[] = "DWRSXP";
static const char *dftp_defaultflagnames[] = { "Delete","Upload","Download","Siteop","Extended WHO","Password change" };

/* Conversion functions */

struct ident_ip *dftp_text2identip(char *src, int *count)
{
   char *ptr = src,*ptr2,*ptr3; struct ident_ip *idip = NULL, *idptr;
   int i;

   *count = 0;
   while((ptr2 = strsep(&ptr," ")) && strlen(ptr2))
   {
      ptr3 = ptr2;
      idip = realloc(idip,++(*count)*sizeof(struct ident_ip));
      idptr = &idip[*count-1];
      memset(idptr,0,sizeof(struct ident_ip));
      if(strchr(ptr2,'!'))
      {
         idptr->flags = 0;
         ptr3 = strsep(&ptr2,"!");
         for(i=0;i<strlen(ptr3);i++)
            switch(ptr3[i])
            {
               case 'C': idptr->flags |= ID_CONTROL; break;
               case 'D': idptr->flags |= ID_DATA; break;
               case 'B': idptr->flags |= ID_BAN; break;
            }
         ptr3 = ptr2;
      }
      if(strchr(ptr2,'@'))
      {
         ptr3 = strsep(&ptr2,"@"); idptr->ident = malloc(strlen(ptr3)+1);
         strcpy(idptr->ident,ptr3); ptr3 = strsep(&ptr2,"@");
      }
      idptr->ip = malloc(strlen(ptr3)+1); strcpy(idptr->ip,ptr3);
   }
   return idip;
}

char *dftp_identip2text(const struct ident_ip *idip,const  int count)
{
   char *output = NULL, flags[5]; int size,i;

   if(!idip || !idip->ip) return NULL;
   for(i=0,size=0;i<count;i++)
   {
      if(i)
      {
         output = realloc(output,++size+1); strcat(output," ");
      }
      memset(flags,0,5);
      if(idip[i].flags)
      {
         if(idip[i].flags&ID_CONTROL) strcat(flags,"C");
         if(idip[i].flags&ID_DATA) strcat(flags,"D");
         if(idip[i].flags&ID_BAN) strcat(flags,"B");
         strcat(flags,"!");
         size += strlen(flags);
         if(!output)
         {
            output = malloc(size+1); memset(output,0,1);
         }
         output = realloc(output,size+1); strcat(output,flags);
      }
      if(idip[i].ident)
      {
         size += strlen(idip[i].ident)+1;
         if(!output)
         {
            output = malloc(size+1); memset(output,0,1);
         }
         output = realloc(output,size+1);
         strcat(output,idip[i].ident); strcat(output,"@");
      }
      if(idip[i].ip)
      {
         size += strlen(idip[i].ip);
         if(!output)
         {
            output = malloc(size+1); memset(output,0,1);
         }
         output = realloc(output,size+1);
         strcat(output,idip[i].ip);
      }
   }
   return output;
}

/* Configuration functions */

int dftp_getconfline(const int fd, char **token, char **value)
{
   int i; char *rivi = NULL, *ptr;

   *token = NULL; *value = NULL;
   rivi = dftp_readln(fd);
   if(!rivi || !strlen(rivi)) return 2;

   if(strchr(rivi,'#')) *(strchr(rivi,'#')) = 0;
   if(strchr(rivi,'\n')) *(strchr(rivi,'\n')) = 0;
   if(strchr(rivi,'=')) ptr = strchr(rivi,'=');
   else
   {
      free(rivi); return 1;
   }
   *ptr++ = 0; 
   for(i = strlen(rivi)-1; i >= 0; i--)
   {
      if(isspace(rivi[i])) rivi[i] = 0;
      else break;
   }
   if(!strlen(rivi))
   {
      free(rivi); return 1;
   }
   *token = malloc(strlen(rivi)+1); strcpy(*token,rivi);

   for(;strlen(ptr);ptr++)
   {
      if(isspace(*ptr)) *ptr = 0;
      else break;
   }
   for(i = strlen(ptr)-1; i >= 0; i--)
   {
      if(isspace(ptr[i])) ptr[i] = 0;
      else break;
   }
   if(!strlen(ptr))
   {
      free(rivi); return 1;
   }
   *value = malloc(strlen(ptr)+1); strcpy(*value,ptr);
   return 0;
}

int dftp_readconfig(struct dftp_config *config)
{
   int i,result,fd = -1;
   char *string = NULL, *token = NULL, *value = NULL;

   /* Reset to default values */

   memset(config,0,sizeof(struct dftp_config));
   config->maxviewsize = 10*KB; config->cmode = 0644;
   config->loginidletime = 1; config->maxidletime = 3;
   config->stupdatetime = 10;
   config->sitedir = malloc(strlen(site_defaultpath)+1);
   strcpy(config->sitedir,site_defaultpath);
   for(i=0;i<strlen(dftp_defaultflags);i++)
   {
      config->flags = realloc(config->flags,++config->flagcount*sizeof(struct dftp_flag));
      config->flags[i].mark = dftp_defaultflags[i];
      config->flags[i].name = malloc(strlen(dftp_defaultflagnames[i])+1);
      strcpy(config->flags[i].name,dftp_defaultflagnames[i]);
   }
   
   /* Get locationg of dftpd.conf */

   if(!(string = getenv(DFTP_ENV)))
   {
      config->confdir = malloc(strlen(conf_defaultpath)+1);
      strcpy(config->confdir,conf_defaultpath);
   }
   else
   {
      config->confdir = malloc(strlen(string)+1);
      strcpy(config->confdir,string);
      if(config->confdir[strlen(config->confdir)-1] == '/')
         config->confdir[strlen(config->confdir)-1] = 0;
   }
   string = malloc(strlen(config->confdir)+strlen(CONFIGFILE)+2);
   sprintf(string,"%s/%s",config->confdir,CONFIGFILE);
   fd = open(string,O_RDONLY); free(string);
   if(fd < 0) return 0;

   flock(fd,LOCK_EX);
   for(;;)
   {
      result = dftp_getconfline(fd,&token,&value);
      if(result == 2)
      {
         free(token); free(value);
         if(!config->httpport) config->httpport = config->port+1;
         flock(fd,LOCK_UN); return 1;
      }
      if(result < 2 && token && !strncmp(token,"log ",4))
      {
         config->logs = realloc(config->logs,++config->logcount*sizeof(struct dftp_log));
         memset(&config->logs[config->logcount-1],0,sizeof(struct dftp_log));
         config->logs[config->logcount-1].pattern = malloc(strlen(token)-4+1);
         strcpy(config->logs[config->logcount-1].pattern,token+4);
         if(value)
         {
            config->logs[config->logcount-1].path = malloc(strlen(value)+1);
            strcpy(config->logs[config->logcount-1].path,value);
         }
      }
      if(!result)
      {
         if(!strncmp(token,"ulcheck ",8))
         {
            config->ulchecker = realloc(config->ulchecker,++config->uccount*sizeof(struct dftp_ulchecker));
            config->ulchecker[config->uccount-1].mask = malloc(strlen(token)-8+1);
            config->ulchecker[config->uccount-1].handler = malloc(strlen(value)+1);
            strcpy(config->ulchecker[config->uccount-1].mask,token+8);
            strcpy(config->ulchecker[config->uccount-1].handler,value);
         }
         if(!strcmp(token,"dupecheck"))
         {
            config->dupechecker = realloc(config->dupechecker,++config->dccount*sizeof(void *));
            config->dupechecker[config->dccount-1] = malloc(strlen(value)+1);
            strcpy(config->dupechecker[config->dccount-1],value);
         }
         if(!strcmp(token,"port")) config->port = atoi(value);
         if(!strcmp(token,"httpport")) config->httpport = atoi(value);
         if(!strcmp(token,"maxloginfails")) config->maxloginfails = atoi(value);
         if(!strcmp(token,"maxidletime")) config->maxidletime = atoi(value);
         if(!strcmp(token,"loginidletime")) config->loginidletime = atoi(value);
         if(!strcmp(token,"stupdatetime")) config->stupdatetime = atoi(value);
         if(!strcmp(token,"debuglevel")) config->debuglevel = atoi(value);
         if(!strcmp(token,"createmode")) sscanf(value,"%o",&config->cmode);
         if(!strcmp(token,"welcomefile") && config->welcomefile == NULL)
         {
            config->welcomefile = malloc(strlen(value)+1);
            strcpy(config->welcomefile,value);
         }
         if(!strcmp(token,"sitedir"))
         {
            if(value[strlen(value)-1] == '/') value[strlen(value)-1] = 0;
            config->sitedir = realloc(config->sitedir,strlen(value)+1);
            strcpy(config->sitedir,value);
         }
         if(!strcmp(token,"welcomemsg") && config->welcomemsg == NULL)
         {
            config->welcomemsg = malloc(strlen(value)+1);
            strcpy(config->welcomemsg,value);
         }
         if(!strcmp(token,"goodbyemsg") && config->goodbyemsg == NULL)
         {
            config->goodbyemsg = malloc(strlen(value)+1);
            strcpy(config->goodbyemsg,value);
         }
         if(!strcmp(token,"statusline") && config->statusline == NULL)
         {
            config->statusline = malloc(strlen(value)+1);
            strcpy(config->statusline,value);
         }
         if(!strcmp(token,"proctitle")) config->mod_proctitle = value[0];
         if(!strcmp(token,"resolvehosts")) config->resolvehosts = value[0];
         if(!strcmp(token,"realusersupport")) config->realusersupport = value[0];
         if(!strcmp(token,"shadow")) config->shadow = value[0];
      }
      free(token); token = NULL;
      free(value); value = NULL;
   }
}

void dftp_freeconfig(struct dftp_config *cnf)
{
   if(cnf->confdir) free(cnf->confdir);
   if(cnf->sitedir) free(cnf->sitedir);
   for(;cnf->logcount;cnf->logcount--)
   {
      free(cnf->logs[cnf->logcount-1].pattern);
      free(cnf->logs[cnf->logcount-1].path);
   }
   free(cnf->logs);
   for(;cnf->uccount;cnf->uccount--)
   {
      free(cnf->ulchecker[cnf->uccount-1].mask);
      free(cnf->ulchecker[cnf->uccount-1].handler);
   }
   free(cnf->ulchecker);
   for(;cnf->dccount;cnf->dccount--) free(cnf->dupechecker[cnf->dccount-1]);
   free(cnf->dupechecker);
   if(cnf->welcomefile) free(cnf->welcomefile);
   if(cnf->welcomemsg) free(cnf->welcomemsg);
   if(cnf->goodbyemsg) free(cnf->goodbyemsg);
   if(cnf->statusline) free(cnf->statusline);
   
}
/* User database functions */

void dftp_uaddent(const int fd, struct dftp_uent *entry)
{
   char teksti[30];

   if(lseek(fd,-1,SEEK_CUR) >= 0)
   {
      if(read(fd,teksti,1) < 0) return;
      if(teksti[0] != '\n') write(fd,"\n",1);
   }

   write(fd,entry->name,strlen(entry->name)); write(fd,":",1);
   write(fd,entry->passwd,strlen(entry->passwd)); write(fd,":",1);
   if(entry->flags)
      write(fd,entry->flags,strlen(entry->flags)); write(fd,":",1);

   sprintf(teksti,"%i",entry->maxconn);
   write(fd,teksti,strlen(teksti)); write(fd,":",1);
   sprintf(teksti,"%u",entry->uid);
   write(fd,teksti,strlen(teksti)); write(fd,":",1);
   sprintf(teksti,"%u",entry->gid);
   write(fd,teksti,strlen(teksti)); write(fd,":",1);
   sprintf(teksti,"%i",entry->ratio);
   write(fd,teksti,strlen(teksti)); write(fd,":",1);
   sprintf(teksti,"%qu",entry->ulbytes);
   write(fd,teksti,strlen(teksti)); write(fd,":",1);
   sprintf(teksti,"%qu",entry->dlbytes);
   write(fd,teksti,strlen(teksti)); write(fd,":",1);
   sprintf(teksti,"%qu",entry->credits);
   write(fd,teksti,strlen(teksti)); write(fd,":",1);
   if(entry->homedir) write(fd,entry->homedir,strlen(entry->homedir));
   write(fd,":",1);
   if(entry->comment) write(fd,entry->comment,strlen(entry->comment));
   write(fd,":",1);
   if(entry->identip)
   {
      char *text = dftp_identip2text(entry->identip,entry->identipcount);

      write(fd,text,strlen(text)); free(text);
   }
   write(fd,"\n",1);
}

void dftp_uaddsiteent(const int fd, struct dftp_uent *entry)
{
   char teksti[20];

   if(lseek(fd,-1,SEEK_CUR) >= 0)
   {
      if(read(fd,teksti,1) < 0) return;
      if(teksti[0] != '\n') write(fd,"\n",1);
   }

   write(fd,entry->name,strlen(entry->name)); write(fd,":*:",3);
   sprintf(teksti,"%u",entry->uid);
   write(fd,teksti,strlen(teksti)); write(fd,":",1);
   sprintf(teksti,"%u",entry->gid);
   write(fd,teksti,strlen(teksti)); write(fd,"::",2);
   if(entry->homedir) write(fd,entry->homedir,strlen(entry->homedir));
   else write(fd,"/",1); write(fd,":\n",2);
}

int dftp_ugetent(const int fd, struct dftp_uent *entry)
{
   int i;
   char c, *line = NULL, *ptr, *ptr2;

   i = read(fd,&c,1);
   if(i < 1 || !isprint(c)) return 0;
   lseek(fd, -1, SEEK_CUR);

   memset(entry,0,sizeof(struct dftp_uent));
   line = dftp_readln(fd);
   ptr = line;
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   strncpy(entry->name,ptr2,NAMELEN);
   entry->name[NAMELEN] = '\0';
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   strncpy(entry->passwd,ptr2,PASSWDLEN);
   entry->passwd[PASSWDLEN] = '\0';
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   entry->flags = malloc(strlen(ptr2)+1);
   strcpy(entry->flags,ptr2);
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   entry->maxconn = atoi(ptr2);
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   entry->uid = atoi(ptr2);
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   entry->gid = atoi(ptr2);
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   entry->ratio = atoi(ptr2);
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   sscanf(ptr2,"%qu",&entry->ulbytes);
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   sscanf(ptr2,"%qu",&entry->dlbytes);
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   sscanf(ptr2,"%qu",&entry->credits);
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   if(strlen(ptr2))
   {
      entry->homedir = malloc(strlen(ptr2)+1); strcpy(entry->homedir,ptr2);
   }
   ptr2 = strsep(&ptr,":"); if(ptr2 == NULL) { free(line); return 1; }
   if(strlen(ptr2))
   {
      entry->comment = malloc(strlen(ptr2)+1); strcpy(entry->comment,ptr2);
   }
   ptr2 = strsep(&ptr,":");
   if(ptr2 != NULL)
      entry->identip = dftp_text2identip(ptr2,&entry->identipcount);
   free(line);
   return(-1);
}

void dftp_ufree(struct dftp_uent *entry)
{
   int i;

   for(i=0;i<entry->identipcount;i++)
   {
      free(entry->identip[i].ident);
      free(entry->identip[i].ip);
   }
   free(entry->flags); free(entry->homedir); free(entry->comment);
   free(entry->identip); memset(entry,0,sizeof(struct dftp_uent));
}

int dftp_ugetbyname(const int fd, char *name, struct dftp_uent *entry)
{
   int i; struct dftp_uent searchentry;

   lseek(fd,0,SEEK_SET);
   while( (i=dftp_ugetent(fd,&searchentry)) )
   {
      if(i < 0 && !strcmp(searchentry.name,name))
      {
         memcpy(entry,&searchentry,sizeof(struct dftp_uent));
         return -1;
      }
      else dftp_ufree(&searchentry);
   }
   return 0;
}

int dftp_ugetbyid(const int fd, unsigned int uid, struct dftp_uent *entry)
{
   int i; struct dftp_uent searchentry;

   lseek(fd,0,SEEK_SET);
   while( (i=dftp_ugetent(fd,&searchentry)) )
   {
      if(i < 0 && searchentry.uid == uid)
      {
         memcpy(entry,&searchentry,sizeof(struct dftp_uent));
         return -1;
      }
      else dftp_ufree(&searchentry);
   }
   return 0;
}

void dftp_uupdatedb(const int fd, struct dftp_uent *entry)
{
   struct dftp_uent uuso, *users = NULL; int i, ucount = 0;
   off_t offset = 0;

   if(fd < 0) return; lseek(fd,0,SEEK_SET);
   /* Read users until target is found */
   while((i = dftp_ugetent(fd,&uuso)) && uuso.uid != entry->uid)
   {
       offset = lseek(fd,0,SEEK_CUR); dftp_ufree(&uuso);
   }
   if(!i) return; /* target user not found */
   while((i=dftp_ugetent(fd,&uuso))) /* read rest of the users */
      if(i < 0)
      {
         users = realloc(users,++ucount*sizeof(struct dftp_uent));
         memcpy(&users[ucount-1],&uuso,sizeof(struct dftp_uent));
      }
   lseek(fd,offset,SEEK_SET);
   dftp_uaddent(fd,entry); /* update target user */
   for(i=0;i<ucount;i++) dftp_uaddent(fd,&users[i]); /* others */
   ftruncate(fd,lseek(fd,0,SEEK_CUR)); /* no extra garbage at end of file */
   free(users);
}

void dftp_uupdatesitedb(const int fd, int sitefd)
{
   int i; struct dftp_uent buffer;

   lseek(fd,0,SEEK_SET);
   lseek(sitefd,0,SEEK_SET);

   write(sitefd,"root:*:0:0:::\n",strlen("root:*:0:0:::\n"));
   while((i=dftp_ugetent(fd,&buffer)))
   {
      if(i < 0) dftp_uaddsiteent(sitefd,&buffer);
      dftp_ufree(&buffer);
   }
   ftruncate(sitefd,lseek(sitefd,0,SEEK_CUR));
}

int dftp_udelent(const int fd, unsigned int uid)
{
   struct dftp_uent entry; int result; off_t wr_offset = 0, r_offset = 0;

   lseek(fd,0,SEEK_SET);
   do
   {
      wr_offset = lseek(fd,0,SEEK_CUR);
      result = dftp_ugetent(fd,&entry); r_offset = lseek(fd,0,SEEK_CUR);
      if(result && entry.uid == uid) result = 2; dftp_ufree(&entry);
   } while(result && result != 2);
   if(!result) return 0;
   while((result = dftp_ugetent(fd,&entry)))
   {
      r_offset = lseek(fd,0,SEEK_CUR);
      if(result == -1)
      {
         lseek(fd,wr_offset,SEEK_SET);
         dftp_uaddent(fd,&entry); dftp_ufree(&entry);
         wr_offset = lseek(fd,0,SEEK_CUR); lseek(fd,r_offset,SEEK_SET);
      }
   }
   lseek(fd,wr_offset,SEEK_SET); ftruncate(fd,lseek(fd,0,SEEK_CUR));
   return 1;
}

/* Group data base functions */

void dftp_gaddent(const int fd, struct dftp_gent *entry)
{
   char teksti[30]; int i;

   if(lseek(fd,-1,SEEK_CUR) >= 0)
   {
      if(read(fd,teksti,1) < 0) return;
      if(teksti[0] != '\n') write(fd,"\n",1);
   }

   write(fd,entry->name,strlen(entry->name)); write(fd,":",1);
   sprintf(teksti,"%i",entry->id);
   write(fd,teksti,strlen(teksti)); write(fd,":",1);
   for(i=0;i<entry->membercount;i++)
   {
      if(i>0) write(fd,",",1);
      write(fd,entry->members[i],strlen(entry->members[i]));
   }
   write(fd,"\n",1);
}

void dftp_gaddsiteent(const int fd, struct dftp_gent *entry)
{
   char teksti[20]; int i;

   if(lseek(fd,-1,SEEK_CUR) >= 0)
   {
      if(read(fd,teksti,1) < 0) return;
      if(teksti[0] != '\n') write(fd,"\n",1);
   }

   write(fd,entry->name,strlen(entry->name)); write(fd,"::",2);
   sprintf(teksti,"%i",entry->id);
   write(fd,teksti,strlen(teksti)); write(fd,":",1);
   for(i=0;i<entry->membercount;i++)
   {
      if(i>0) write(fd,",",1);
      write(fd,entry->members[i],strlen(entry->members[i]));
   }
   write(fd,"\n",1);
}

int dftp_ggetent(const int fd, struct dftp_gent *entry)
{
   int i; 
   char c, *line = NULL, *ptr, *ptr2;

   i = read(fd,&c,1);
   if(i < 1 || !isprint(c)) return 0;
   lseek(fd, -1, SEEK_CUR);

   entry->members = NULL; entry->membercount = 0;
   line = dftp_readln(fd);
   ptr2 = line;
   ptr = strsep(&ptr2,":"); if(ptr == NULL) { free(line); return 1; }
   strncpy(entry->name,ptr,NAMELEN);
   entry->name[NAMELEN] = '\0';
   ptr = strsep(&ptr2,":"); if(ptr == NULL) { free(line); return 1; }
   entry->id = atoi(ptr);
   ptr = strsep(&ptr2,":");
   ptr2 = ptr;
   while( (ptr = strsep(&ptr2,",")) != NULL)
   {
      if(strlen(ptr) == 0) break;
      entry->members = realloc(entry->members,++entry->membercount*sizeof(char *));
      entry->members[entry->membercount-1] = malloc(strlen(ptr)+1);
      strcpy(entry->members[entry->membercount-1],ptr);
   }
   free(line); return(-1);
}

void dftp_gfree(struct dftp_gent *entry)
{
   int i;

   for(i=0;i<entry->membercount;i++) free(entry->members[i]);
   free(entry->members); memset(entry,0,sizeof(struct dftp_gent));
}

int dftp_ggetbyname(const int fd, char *name, struct dftp_gent *entry)
{
   int i; struct dftp_gent searchentry;

   lseek(fd,0,SEEK_SET);
   while( (i=dftp_ggetent(fd,&searchentry)) )
   {
      if(i < 0 && !strcmp(searchentry.name,name)) 
      {
         memcpy(entry,&searchentry,sizeof(struct dftp_gent));
         return -1;
      }
      else dftp_gfree(&searchentry);
   }
   return 0;
}

int dftp_ggetbyid(const int fd, unsigned int id, struct dftp_gent *entry)
{
   int i; struct dftp_gent searchentry;

   lseek(fd,0,SEEK_SET);
   while( (i=dftp_ggetent(fd,&searchentry)) )
   {
      if(i < 0 && searchentry.id == id) 
      {
         memcpy(entry,&searchentry,sizeof(struct dftp_gent));
         return -1;
      }
      else dftp_gfree(&searchentry);
   }
   return 0;
}

void dftp_gupdatedb(const int fd, struct dftp_gent *entry)
{  
   struct dftp_gent ryhma, *groups = NULL; int i, gcount = 0;
   off_t offset = 0;

   if(fd < 0) return;
   lseek(fd,0,SEEK_SET);
   while((i = dftp_ggetent(fd,&ryhma)) && ryhma.id != entry->id)
   {
       offset = lseek(fd,0,SEEK_CUR); dftp_gfree(&ryhma);
   }
   if(!i) return; /* target group not found */
   while((i=dftp_ggetent(fd,&ryhma))) /* read rest of the groups */
      if(i < 0)
      {
         groups = realloc(groups,++gcount*sizeof(struct dftp_gent));
         memcpy(&groups[gcount-1],&ryhma,sizeof(struct dftp_gent));
      }
   lseek(fd,offset,SEEK_SET);
   dftp_gaddent(fd,entry); /* update target group */
   for(i=0;i<gcount;i++) dftp_gaddent(fd,&groups[i]); /* others */
   ftruncate(fd,lseek(fd,0,SEEK_CUR)); /* no extra garbage at end of file */
   free(groups);
}

void dftp_gupdatesitedb(const int fd, int sitefd)
{
   int i; struct dftp_gent buffer;

   lseek(fd,0,SEEK_SET);
   lseek(sitefd,0,SEEK_SET);

   write(sitefd,"root::0:root\n",strlen("root::0:root\n"));
   while((i=dftp_ggetent(fd,&buffer)))
   {
      if(i < 0) dftp_gaddsiteent(sitefd,&buffer);
      dftp_gfree(&buffer);
   }

   ftruncate(sitefd,lseek(sitefd,0,SEEK_CUR));
}

int dftp_gdelent(const int fd, unsigned int id)
{
   struct dftp_gent entry; int result; off_t wr_offset = 0, r_offset = 0;

   lseek(fd,0,SEEK_SET);
   do
   {
      memset(&entry,0,sizeof(struct dftp_gent));
      wr_offset = lseek(fd,0,SEEK_CUR);
      result = dftp_ggetent(fd,&entry); r_offset = lseek(fd,0,SEEK_CUR);
      if(result && entry.id == id) result = 2; dftp_gfree(&entry);
   } while(result && result != 2);
   if(!result) return 0;
   while((result = dftp_ggetent(fd,&entry)))
   {
      r_offset = lseek(fd,0,SEEK_CUR);
      if(result == -1)
      {
         lseek(fd,wr_offset,SEEK_SET);
         dftp_gaddent(fd,&entry); dftp_gfree(&entry);
         wr_offset = lseek(fd,0,SEEK_CUR); lseek(fd,r_offset,SEEK_SET);
      }
   }
   lseek(fd,wr_offset,SEEK_SET); ftruncate(fd,lseek(fd,0,SEEK_CUR));
   return 1;
}

/* UTMP functions */

void dftp_utmpaddent(const int fd, struct dftp_utmpent *entry)
{
   char teksti[NAMELEN+ACTIVITYLEN+2*HOSTNAMELEN+33];

   if(lseek(fd,-1,SEEK_CUR) > 0)
   {
      if(read(fd,teksti,1) < 0) return;
      if(teksti[0] != '\n') write(fd,"\n",1);
   }

   memset(teksti,0,sizeof(teksti));
   sprintf(teksti,UTMPSTRING,entry->name,entry->pid,entry->aika,
      entry->idleaika,entry->activity,entry->controlhostname,
      entry->datahostname);
   write(fd,teksti,strlen(teksti));
}

int dftp_utmpgetent(const int fd, struct dftp_utmpent *entry)
{
   int i; 
   char c, *line = NULL, *ptr, *ptr2;

   i = read(fd,&c,1);
   if(i < 1 || (!isalnum(c) && c != ' ') ) return 0;
   lseek(fd, -1, SEEK_CUR);
   memset(entry,0,sizeof(struct dftp_utmpent));
   line = dftp_readln(fd);
   ptr2 = line;
   ptr = strsep(&ptr2,":"); if(ptr == NULL) { free(line); return 1; }
   strncpy(entry->name,ptr,NAMELEN); 
   entry->name[NAMELEN] = '\0';
   ptr = strsep(&ptr2,":"); if(ptr == NULL) { free(line); return 1; }
   entry->pid = atoi(ptr);
   ptr = strsep(&ptr2,":"); if(ptr == NULL) { free(line); return 1; }
   entry->aika = atoi(ptr);
   ptr = strsep(&ptr2,":"); if(ptr == NULL) { free(line); return 1; }
   entry->idleaika = atoi(ptr);
   ptr = strsep(&ptr2,":"); if(ptr == NULL) { free(line); return 1; }
   strncpy(entry->activity,ptr,ACTIVITYLEN);
   ptr = strsep(&ptr2,":"); if(ptr == NULL) { free(line); return 1; }
   strncpy(entry->controlhostname,ptr,HOSTNAMELEN);
   ptr = strsep(&ptr2,":");
   if(ptr != NULL) strncpy(entry->datahostname,ptr,HOSTNAMELEN);

   free(line);
   return(-1);
}

void dftp_utmpupdate(const int fd, struct dftp_utmpent *entry)
{
   int i; off_t pos = 0;
   struct dftp_utmpent uuso;

   lseek(fd,0,SEEK_SET);
   while( (i=dftp_utmpgetent(fd,&uuso)) )
   {
      if(i < 0 && uuso.pid == entry->pid)
      {
         lseek(fd,pos,SEEK_SET);
         dftp_utmpaddent(fd,entry); return;
      }
      pos = lseek(fd,0,SEEK_CUR);
   }
   dftp_utmpaddent(fd,entry);
}

void dftp_utmpdelent(const int fd, int pid)
{  
   int i,lkm = 0;
   struct dftp_utmpent **buffer = NULL; struct dftp_utmpent uuso;

   lseek(fd,0,SEEK_SET);
   while( (i=dftp_utmpgetent(fd,&uuso)) )
   {
      if(i < 0 && uuso.pid != pid)
      {
         buffer = realloc(buffer,(lkm+1)*sizeof(void *));
         buffer[lkm] = malloc(sizeof(struct dftp_utmpent));
         memcpy(buffer[lkm++],&uuso,sizeof(struct dftp_utmpent));
      }
   }
   lseek(fd,0,SEEK_SET);
   for(i = 0; i < lkm; i++) dftp_utmpaddent(fd,buffer[i]);
   free(buffer); ftruncate(fd,lseek(fd,0,SEEK_CUR));
}

/* Dupe Checker functions */

int dftp_dupeaddent(const int fd, struct dftp_dupeent *entry)
{
   char teksti[20]; int koko = 0;

   if(lseek(fd,-1,SEEK_CUR) >= 0)
   {
      if(read(fd,teksti,1) < 0) return 0;
      if(teksti[0] != '\n') write(fd,"\n",1);
   }

   if(entry->filename != NULL) koko += write(fd,entry->filename,strlen(entry->filename));
   koko += write(fd,":",1);

   sprintf(teksti,"%li",(long int)entry->ultime);
   koko += write(fd,teksti,strlen(teksti)); koko += write(fd,":",1);

   koko += write(fd,entry->uploader,strlen(entry->uploader));
   koko += write(fd,":",1);

   if(entry->dirname != NULL) koko += write(fd,entry->dirname,strlen(entry->dirname));
   koko += write(fd,"\n",1);

   return(koko);
}

int dftp_dupegetent(const int fd, struct dftp_dupeent *entry)
{
   int i;
   char c, *line = NULL, *ptr, *ptr2;

   i = read(fd,&c,1);
   if(i < 1 || !isalnum(c)) return 0;
   lseek(fd, -1, SEEK_CUR);

   line = dftp_readln(fd);
   ptr2 = line;
   ptr = strsep(&ptr2,":");
   entry->filename = malloc(strlen(ptr)+1); strcpy(entry->filename,ptr);
   ptr = strsep(&ptr2,":"); entry->ultime = (time_t)atol(ptr);
   ptr = strsep(&ptr2,":"); strncpy(entry->uploader,ptr,NAMELEN);
   entry->uploader[NAMELEN] = '\0';
   ptr = strsep(&ptr2,":");
   entry->dirname = malloc(strlen(ptr)+1); strcpy(entry->dirname,ptr);
   free(line);
   return(-1);
}

/* Other functions */

char *dftp_readln(const int fd)
{
   int i,koko = 0; char tempid[30], *line = NULL, *ptr;

   while( (i = read(fd,tempid,sizeof(tempid))) > 0)
   {
      koko += i; line = realloc(line,koko); memcpy(line+koko-i,tempid,i);
      if( (ptr = memchr(tempid,'\n',i)) != NULL)
      {
         koko = koko-i+(ptr-tempid); line = realloc(line,koko);
         lseek(fd,-(i-(ptr-tempid))+1,SEEK_CUR); break;
      }
   }
   line = realloc(line,koko+1); line[koko] = '\0';
   return(line);
}

char *dftp_printtime(time_t time1)
{
   time_t time2; int tot_secs,mins = 0,hrs = 0,days = 0;
   static char value[25];

   time(&time2); tot_secs = (int)difftime(time2,time1);
   days = tot_secs/84600; tot_secs -= days*86400;
   hrs = tot_secs/3600; tot_secs -= hrs*3600;
   mins = tot_secs/60; tot_secs -= mins*60;
   memset(value,0,25);
   if(days) sprintf(value,"%id ",days);
   if(hrs) sprintf(value+strlen(value),"%ih ",hrs);
   if(mins) sprintf(value+strlen(value),"%im ",mins);
   if(tot_secs || (!days && !hrs && !mins)) sprintf(value+strlen(value),"%is ",tot_secs);
   return value;
}

char *dftp_printbytes(unsigned long long bytes)
{
   static char luku[30];

   if(bytes >= GB)
   {
      sprintf(luku,"%.1f",(float)(bytes/(float)GB));
      if(*(strchr(luku,'.')+1) != '0') *(strchr(luku,'.')+2) = 0;
      else *strchr(luku,'.') = 0;
      strcat(luku,"Gb");
   }
   else if(bytes >= MB)
   {
      sprintf(luku,"%.1f",(float)(bytes/(float)MB));
      if(*(strchr(luku,'.')+1) != '0') *(strchr(luku,'.')+2) = 0;
      else *strchr(luku,'.') = 0;
      strcat(luku,"Mb");
   }
   else if(bytes >= KB)
   {
      sprintf(luku,"%.1f",(float)(bytes/(float)KB));
      if(*(strchr(luku,'.')+1) != '0') *(strchr(luku,'.')+2) = 0;
      else *strchr(luku,'.') = 0;
      strcat(luku,"Kb");
   }
   else sprintf(luku,"%u",(unsigned int)bytes);
   return luku;
}

void dftp_creds(char *args, unsigned long long *creds)
{
   int direction = 0; unsigned long long kerroin = 1;

   if(args == NULL || creds == NULL) return;
   if(toupper(args[strlen(args)-1]) == 'K') kerroin = KB;
   else if(toupper(args[strlen(args)-1]) == 'M') kerroin = MB;
   else if(toupper(args[strlen(args)-1]) == 'G') kerroin = GB;
   if(args[0] == '+') direction = 1;
   else if(args[0] == '-') direction = -1;
   if(direction)
   {
      if(direction == -1 && *creds < atoi(args+1)*kerroin) *creds = 0;
      else *creds += atoi(args+1)*direction*kerroin;
   }
   else
   {
      sscanf(args,"%qu",creds); *creds *= kerroin;
   }
}

int dftp_isnumeric(char *string)
{
   int i;

   for(i=0;i<strlen(string);i++)
      if(isalpha(string[i])) return 0;
   return 1;
}
