#include <stdio.h>
#include <stdlib.h>
#ifndef __USE_GNU
#define __USE_GNU
#endif
#ifndef __USE_XOPEN
#define __USE_XOPEN
#endif
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <time.h>
#include <ctype.h>
#include <dirent.h>
#include <fcntl.h>
#include <glob.h>
#include <fnmatch.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include "../dftpd.h"

#define VIEWDIR "sview"

/* libdftp.c */
int dftp_ugetbyid(int, int, struct dftp_uent *);
void dftp_uupdatedb(int, struct dftp_uent *);
void dftp_ufree(struct dftp_uent *);
int dftp_ggetent(int, struct dftp_gent *);
void dftp_gfree(struct dftp_gent *);
int dftp_ugetbyname(int , char *, struct dftp_uent *);
int dftp_ggetbyname(int , char *, struct dftp_gent *);
int dftp_utmpgetent(int, struct dftp_utmpent *);
int dftp_isnumeric(char *);

void (*writes)(const char *, ...) = NULL;
void (*cwrites)(const char *, ...) = NULL;
void (*ftplog)(const char *, const char *, ...) = NULL;
void (*viewfile)(const char *, const short) = NULL;

const char notloggedtxt[] = "530 Please login with USER and PASS.\r\n";

struct dftp_config *cfg; struct dftp_vars *var;
int passwdfile = -1, groupfile = -1, utmpfile = -1, viewdir = -1, sockfd = -1;

void initplugin(struct dftp_plugininfo pinfo)
{
   char filename[PATH_MAX];

   writes = pinfo.writes; cwrites = pinfo.cwrites; ftplog = pinfo.ftplog; viewfile = pinfo.viewfile;
   cfg = pinfo.cfg; var = pinfo.var;
   passwdfile = pinfo.passwdfile; groupfile = pinfo.groupfile; utmpfile = pinfo.utmpfile; sockfd = pinfo.sockfd;

   /* Initialize SITE VIEW(cmd #4) */

   sprintf(filename,"%s/%s",cfg->confdir,VIEWDIR);
   viewdir = open(filename,0);
   if(viewdir < 0) ftplog("ERR","Failed to open site view directory");
}

/* SITE EXEC */

char **parsestr(char *source)
{
   char **dest = NULL, *ptr = source, *ptr2; int destc = 0;

   while((ptr2=strsep(&ptr," ")))
   {
      dest = realloc(dest,(++destc+1)*sizeof(void *));
      dest[destc-1] = malloc(strlen(ptr2)+1); strcpy(dest[destc-1],ptr2);
      dest[destc] = NULL;
   }
   return dest;
}

const char cmdname1[] = "EXEC";
const char cmdhelp1[] = "214 SITE EXEC <command> : execute an external command\r\n";
const char cmdavail1  = ALL;

void cmd1(char *args)
{
   int pipefd[2], cpid, nleft, nwritten = 0;
   char buf[2], *bufptr;

   if(var->logged < 2)
   {
      writes(notloggedtxt); return;
   }

   if(args == NULL)
   {
      writes(cmdhelp1); return;
   }
   writes("230-Executing '%s'\r\n",args);

   pipe(pipefd);
   if((cpid = fork()) < 0)
   {
      writes("230 Fork error, aborting execution\r\n"); return;
   }
   if(cpid == 0)
   {
      char **cmd = NULL, *text = malloc(strlen(args)+strlen("/bin/exec/")+1);

      dup2(pipefd[1],1); dup2(pipefd[1],2); close(pipefd[1]);
      close(0); for(nleft=3;nleft<20;nleft++) close(nleft);

      sprintf(text,"/bin/exec/%s",args);
      cmd = parsestr(text); execv(cmd[0],cmd); exit(0);
   }
   close(pipefd[1]);

   cpid = 0;
   while(read(pipefd[0],&buf,1) > 0)
   {
      nleft = 1; bufptr = buf;
      if(buf[0] == '\n')
      {
         sprintf(buf,"\r\n "); nleft = 3;
      }
      while(nleft > 0)
      {
         nwritten = write(sockfd,bufptr,nleft);
         if( nwritten < 0) break;

         nleft -= nwritten;
         bufptr += nwritten;
      }
      if( nwritten < 0) break;
   }
   wait(NULL);
   close(pipefd[0]);
   writes("\r\n");
   if(nwritten < 0) 
   {
      ftplog("ERR","write(): %s",strerror(errno));
      writes("230 Error writing to data socket.\r\n");
   }
   else writes("230 Finished execution of '%s'\r\n",args);
}

/* SITE WHO */

const char cmdname2[] = "WHO";
const char cmdhelp2[] = "214 WHO : List connected users, their idletimes and activities\r\n";
const char cmdavail2  = REAL|GUEST;

void cmd2(char *args)
{
   int i,minutes = 0,seconds = 0,count = 0,uusot=0; 
   time_t aika; struct tm *aika2;
   double aikaero; struct dftp_utmpent entry;

   time(&aika); flock(utmpfile,LOCK_EX); lseek(utmpfile,0,SEEK_SET);
   while( (i=dftp_utmpgetent(utmpfile,&entry)) )
      if(i < 0)
      {
         uusot++;
         if(count > 0)
         {
            writes("210---------------------------------------------------------");
            writes("--------------------\r\n");
         }
         aika2 = localtime(&entry.aika);
         if(!strcmp(entry.name,"ftp")) writes("210-User: <anonymous>     ");
         else if(strlen(entry.name) > 0) writes("210-User: %-16s",entry.name);
         else writes("210-User: <NULL>           ");
         if(strchr(var->user.flags,'X')) writes("Pid: %i ",entry.pid);
         writes("Online since: %02i:%02i\r\n",aika2->tm_hour,aika2->tm_min);
         if(strlen(entry.activity) > 0) writes("210-Activity: %s\r\n",entry.activity);
         else
         {
            aikaero = difftime(aika,entry.idleaika);
            minutes = ((int)(aikaero))/60; seconds = ((int)(aikaero))%60;
            writes("210-Activity: idle(%i min %i sec)\r\n",minutes,seconds);
         }
         if(strchr(var->user.flags,'X'))
         {
            if(strlen(entry.controlhostname) > 0) writes("210-Control host name: %s\r\n",entry.controlhostname);
            else writes("210-Control host name: <NULL>\n");
            if(!isspace(entry.datahostname[0])) writes("210-Data host name: %s\r\n",entry.datahostname);
         }
         count++;
      }
   flock(utmpfile,LOCK_UN); writes("210 Total %i user",uusot);
   if(uusot > 1) writes("s"); writes(" online\r\n");
}

/* SITE CHMOD */

const char cmdname3[] = "CHMOD";
const char cmdhelp3[] = "214 CHMOD <mode> <file>: Change access permissions of specified file\r\n";
const char cmdavail3  = REAL|GUEST;

void cmd3(char *args)
{
   unsigned int mode, i; char *ptr = args,*ptr2; glob_t glt;

   if(var->userclass == ANONYMOUS)
   {
      writes("530 Access denied for anonymous user.\r\n"); return;
   }
   ptr2 = strsep(&ptr," ");
   if(!ptr2)
   {
      writes("500 'SITE CHMOD %s': command not understood.\r\n",args); return;
   }
   if(!dftp_isnumeric(ptr2))
   {
      writes("500 'SITE CHMOD %s': command not understood.\r\n",args); return;
   }
   sscanf(ptr2,"%o",&mode);
   ptr2 = strsep(&ptr," ");
   if(!ptr2)
   {
      writes("500 'SITE CHMOD %s': command not understood.\r\n",args); return;
   }
   if(!chmod(ptr2,mode))
   {
      writes("200 CHMOD command successful.\r\n"); return;
   }
   if(errno == ENOENT)
   {
      glob(ptr2,0,NULL,&glt);
      for(i=0;i<glt.gl_pathc;i++) chmod(glt.gl_pathv[i],mode);
      if(glt.gl_pathc)
      {
         writes("200 CHMOD command successful.\r\n"); return;
      }
      else errno = ENOENT;
   }
   writes("550 %s: %s\r\n",ptr2,strerror(errno));
}

/* SITE CHOWN */

int resolveuser(char *name)
{
   int result; struct dftp_uent entry; struct passwd *pwent;

   if(!strcmp(name,"root")) return 0;
   if(var->userclass != REAL)
   {
      flock(passwdfile,LOCK_EX);
      result = dftp_ugetbyname(passwdfile,name,&entry);
      flock(passwdfile,LOCK_UN);
      if(!result) return -1;
      result = entry.uid; dftp_ufree(&entry);
      return result;
   }
   else
   {
      if((pwent = getpwnam(name))) return pwent->pw_uid;
      return -1;
   }
}

int resolvegroup(char *name)
{
   int result; struct dftp_gent entry; struct group *grent;

   if(!strcmp(name,"root")) return 0;
   if(var->userclass != REAL)
   {
      flock(groupfile,LOCK_EX);
      result = dftp_ggetbyname(groupfile,name,&entry);
      flock(groupfile,LOCK_UN);
      if(!result) return -1;
      result = entry.id; dftp_gfree(&entry);
      return result;
   }
   else
   {
      if((grent = getgrnam(name))) return grent->gr_gid;
      return -1;
   }
}

const char cmdname4[] = "CHOWN";
const char cmdhelp4[] = "214 CHOWN [user].[group] <file>: Change owner of specified file\r\n";
const char cmdavail4  = REAL|GUEST;

void cmd4(char *args)
{
   unsigned int uid = -1, gid = -1, i; glob_t glt;
   char *ptr = args, *uidstr = NULL,*gidstr = NULL;

   if(var->userclass == ANONYMOUS)
   {
      writes("530 Access denied for anonymous user.\r\n"); return;
   }
   gidstr = strsep(&ptr," ");
   if(!gidstr)
   {
      writes("500 'SITE CHOWN %s': command not understood.\r\n",args); return;
   }
   uidstr = strsep(&gidstr,".");
   if(!uidstr)
   {
      writes("500 'SITE CHOWN %s': command not understood.\r\n",args); return;
   }
   if(strlen(uidstr))
   {
      if(dftp_isnumeric(uidstr)) uid = atoi(uidstr);
      else if((uid = resolveuser(uidstr)) == -1)
      {
         writes("530 User %s not found.\r\n",uidstr); return;
      }
   }
   if(gidstr)
   {
      if(dftp_isnumeric(gidstr)) gid = atoi(gidstr);
      else if((gid = resolvegroup(gidstr)) == -1)
      {
         writes("530 Group %s not found.\r\n",gidstr); return;
      }
   }
   if(!chown(ptr,uid,gid))
   {
      writes("200 CHOWN command successful.\r\n"); return;
   }
   if(errno == ENOENT)
   {
      glob(ptr,0,NULL,&glt);
      for(i=0;i<glt.gl_pathc;i++) chown(glt.gl_pathv[i],uid,gid);
      if(glt.gl_pathc)
      {
         writes("200 CHOWN command successful.\r\n"); return;
      }
      else errno = ENOENT;
   }
   writes("550 %s: %s\r\n",ptr,strerror(errno));
}

/* SITE VIEW */

const char cmdname5[] = "VIEW";
const char cmdhelp5[] = "214 SITE VIEW <file> : view a textfile over the control connection\r\n";
const char cmdavail5  = ALL;

void cmd5(char *args)
{
   struct stat statbuf; char buf[PATH_MAX], goback = 0;

   if(var->logged < 2)
   {
      writes(notloggedtxt); return;
   }
   if(args == NULL)
   {
      writes(cmdhelp5); return;
   }
   errno = 0;
   if(access(args,F_OK))
   {
      if(strchr(args,'/') == NULL)
      {
         getcwd(buf,512); fchdir(viewdir);
         if(!access(args,F_OK)) goback = 1;
      }
      if(!goback)
      {
         chdir(buf); writes("550 %s: %s.\r\n",args,strerror(ENOENT));
         return;
      }
   }
   stat(args,&statbuf);
   if(cfg->maxviewsize && statbuf.st_size > cfg->maxviewsize)
   {
      writes("550 'SITE VIEW %s': File size exceeds maximum viewable size.\r\n",args);
      return;
   }

   writes("230-Viewing '%s'\r\n ",args);
   viewfile(args,-1); if(goback) chdir(buf);
   writes("230 End of '%s'.\r\n",args);
}

/* SITE FIND */

const char cmdname6[] = "FIND";
const char cmdhelp6[] = "214-FIND <keyword>: Search directory tree case insensitively for keyword\r\n214 FIND -f <UNIX filemask>: Search directory tree by with a UNIX filemask\r\n";
const char cmdavail6  = ALL;

void cmd6(char *args)
{
   char *section = NULL, wd[PATH_MAX+1]; static short maskflag;
   struct stat buf; static unsigned int i,numfiles,level = 0; DIR *dirikka;
   struct dirent *direntry;

   if(++level == 1)
   {
      if(!args)
      {
         writes(cmdhelp6); return;
      }
      maskflag = 0; numfiles = 0;
   }
   section = strsep(&args," ");
   if(section == NULL)
   {
      if(level == 1) writes("505 Required parameter missing.\r\n");
      level--; return;
   }
   for(i=0; i<strlen(section); i++) section[i] = tolower(section[i]);
   if(level == 1)
   {
      if(!strcmp(section,"-f"))
      {
         maskflag = 1;
         section = strsep(&args," ");
         if(section == NULL)
         {
            writes("505 Required parameter missing.\r\n");
            return;
         }
         writes("210-Starting UNIX filemask based search for '%s'\r\n",section);
      }
      else writes("210-Starting keyword based search for '%s'\r\n",section);
   }

   wd[PATH_MAX] = 0; getcwd(wd,PATH_MAX); dirikka = opendir(".");
   while( (direntry = readdir(dirikka)) != NULL && !stat(direntry->d_name,&buf))
      if(strcmp(direntry->d_name,".") && strcmp(direntry->d_name,".."))
      {
         if(!maskflag)
         {
            char *text = malloc(strlen(direntry->d_name)+1);

            strcpy(text, direntry->d_name);
            for(i=0; i<strlen(text); i++) text[i] = tolower(text[i]);

            if(strstr(text,section) != NULL)
            {
               writes("  %s",wd); if(level > 1) writes("/");
               writes("%s\r\n",direntry->d_name);
               numfiles++;
            }
            free(text);
         }
         else
         {
            if(!fnmatch(section,direntry->d_name,FNM_PERIOD|FNM_PATHNAME))
            {
               writes("  %s",wd); if(level > 1) writes("/");
               writes("%s\r\n",direntry->d_name);
               numfiles++;
            }
         }
         if(S_ISDIR(buf.st_mode) && !chdir(direntry->d_name)) cmd5(section);
      }
   closedir(dirikka);

   if(level-- == 1)
   {
      writes("210 %u hit",numfiles);
      if(numfiles != 1) writes("s"); writes(" found\r\n");
   }
   else chdir("..");
}

/* SITE GROUP */

const char cmdname7[] = "GROUP";
const char cmdhelp7[] = "214 SITE GROUP <group name> : change effective group(eGID)\r\n";
const char cmdavail7  = GUEST;

void cmd7(char *args)
{
   int i,maara; struct dftp_gent entry; gid_t lista[NGROUPS_MAX];

   if(args == NULL)
   {
      writes(cmdhelp7); return;
   }
   lseek(groupfile,0,SEEK_SET);
   if( (maara = getgroups(NGROUPS_MAX,lista)) < 0)
   {
      writes("514 getgroups(): %s.\r\n",strerror(errno));
      return;
   }
   while(dftp_ggetent(groupfile,&entry))
   {
      if(!strcmp(entry.name,args))
      {
         for(i = 0; i < maara; i++)
            if(entry.id == lista[i])
            {
               seteuid(0); setegid(entry.id); seteuid(var->user.uid);
               dftp_gfree(&var->group); memcpy(&var->group,&entry,sizeof(struct dftp_gent));
               writes("200 Effective group changed to %s.\r\n",args);
               if(cfg->debuglevel >= 2) ftplog("AUTH","DEBUG: Active group changed to %s",args);
               return;
            }
         writes("500 You are not a member of '%s'.\r\n");
      }
      dftp_gfree(&entry);
   }
   writes("500 Group '%s' not found.\r\n",args);
}

/* SITE GROUPS */

const char cmdname8[] = "GROUPS";
const char cmdhelp8[] = "214 SITE GROUPS : list the groups you are a member of (* => current)\r\n";
const char cmdavail8  = ALL;

void cmd8(char *args)
{
   gid_t lista[NGROUPS_MAX]; int maara = 0,i;
   struct dftp_gent entry; struct group *grent;

   lseek(groupfile,0,SEEK_SET);
   if( (maara = getgroups(NGROUPS_MAX,lista)) < 0)
   {
      writes("514 getgroups(): %s.\r\n",strerror(errno));
      return;
   }
   writes("214 ");
   if(var->userclass != REAL)
      while(dftp_ggetent(groupfile,&entry))
      {
         for(i = 0; i < maara; i++)
            if(entry.id == lista[i])
            {
               if(entry.id == var->user.gid) writes("*"); 
               writes("%s ",entry.name);
            }
      }
   else for(i = 0; i < maara; i++)
      if((grent = getgrgid(lista[i]))) writes("%s ",grent->gr_name);
   writes("\r\n");
}

/* SITE CHPASS */

const char cmdname9[] = "CHPASS";
const char cmdhelp9[] = "214 CHPASS <old password> <new password>: Change your password\r\n";
const char cmdavail9  = GUEST;

void cmd9(char *args)
{
   char *oldpwd,*newpwd;

   if(var->userclass == ANONYMOUS)
   {
      writes("530 Access denied for anonymous user.\r\n"); return;
   }
   if(!strchr(var->user.flags,'P'))
   {
      writes("532 Password change access denied\r\n"); return;
   }

   if(!(oldpwd = strsep(&args," ")) || !(newpwd = strsep(&args," ")))
   {
      writes(cmdhelp9); return;
   }
   if(strcmp(crypt(oldpwd,var->user.passwd),var->user.passwd))
   {
      writes("530 Old password is incorrect.\r\n"); return;
   }
   strcpy(var->user.passwd,crypt(newpwd,var->user.passwd));
   flock(passwdfile,LOCK_EX); dftp_uupdatedb(passwdfile,&var->user);
   flock(passwdfile,LOCK_UN); writes("230 Password changed.\r\n");
}


const char cmdname10[] = "UMASK";
const char cmdhelp10[] = "214 UMASK <numeric mask>: Change create mode to 0777-mask\r\n";
const char cmdavail10  = REAL|GUEST;

void cmd10(char *args)
{
   unsigned int newmode;

   if(!args) writes("200 Current UMASK is %03o\r\n",0777^var->cmode);
   else if(!dftp_isnumeric(args)) writes("500 SITE UMASK %s: command not understood.\r\n");
   else
   {
      if(strlen(args) > 4) writes("501 Bad UMASK value\r\n");
      else
      {
         sscanf(args,"%o",&newmode);
         writes("200 UMASK set to %03o (was %03o)\r\n",newmode,0777^var->cmode);
         var->cmode = 0777^newmode;
      }
   }
}

const char cmdname11[] = "SU";
const char cmdhelp11[] = "214 SU: Switches to and from root access mode\r\n";
const char cmdavail11 = GUEST;

void cmd11(char *args)
{
   if(!strchr(var->user.flags,'S'))
   {
      writes("532 Superuser access denied\r\n"); return;
   }
   if(!geteuid())
   {
      setegid(var->group.id); seteuid(var->user.uid);
      writes("200 Superuser mode is OFF\r\n");
   }   
   else
   {
      seteuid(0); setegid(0); 
      writes("200 Superuser mode is ON\r\n");
   }   
}