#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <fnmatch.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <dirent.h>
#include <ctype.h>
#include <stdarg.h>
#include <string.h>
#include <signal.h>
#include <dlfcn.h>
#include "dftpd.h"

extern char **environ, **Argv; extern int Argc;
extern struct dftp_config cfg;
extern int sockfd, *logfile;
extern struct sockaddr_in cli_addr;
#ifndef HTTPSERV
#define USER var.user
void activity(const char *);
void cwrites(const char *,...);
void viewfile(const char *, const short);
extern int passwdfile, groupfile, utmpfile;
extern struct dftp_vars var;
extern struct sockaddr_in data_addr;
#else
#define USER user
extern struct dftp_uent user;
#endif

/* libdftp.c dependencies */

int dftp_readconfig(struct dftp_config *);
void dftp_freeconfig(struct dftp_config *);

int daemonize()
{
   int childpid, i;

   if( (childpid = fork()) < 0) return(-1);
   else if(childpid > 0) exit(0);

   for(i = 0; i <= 2; i++) close(i); /* Close stdin,stdout,stderr */
   errno = 0; chdir("/"); setsid(); return(0);
}

void ftplog(const char *lclass, const char *fmt, ...)
{
   int i; char buf[384]; time_t aika; struct tm *aika2; va_list ap;

   memset(buf,0,384); time(&aika); aika2 = localtime(&aika);
#ifndef HTTPSERV
   if(var.logged == 2) sprintf(buf,"%s",var.user.name);
#else
   sprintf(buf,"%s",user.name);
#endif
   sprintf(buf+strlen(buf),":%i:%02i%02i%i:%02i%02i%02i:%s:",getpid(),
    aika2->tm_mday,aika2->tm_mon+1,1900+aika2->tm_year,aika2->tm_hour,
    aika2->tm_min,aika2->tm_sec,lclass); va_start(ap,fmt);
   vsnprintf(buf+strlen(buf),382,fmt,ap);
   if(buf[strlen(buf)-1] != '\n') strcat(buf,"\n");

   for(i=0;i<cfg.logcount;i++)
      if(!fnmatch(cfg.logs[i].pattern,buf,FNM_PERIOD))
      {
         if(!cfg.logs[i].path || logfile < 0) return;
         flock(logfile[i],LOCK_EX);
         write(logfile[i],buf,strlen(buf));
         flock(logfile[i],LOCK_UN);
      }
}

void writes(const char *format, ...)
{
   char buf[512]; va_list ap;

   buf[511] = 0; va_start(ap,format); vsnprintf(buf,511,format,ap);
   write(sockfd,buf,strlen(buf));
}

void readline(int socket, char *ptr)
{
   int n,rc;
   char c;

   for(n = 1; n < MAXLINE-1; n++)
   {
      if( (rc = read(socket, &c, 1)) == 1)
      {
         if(c != '\n' && c != '\r') *ptr++ = c;
         if(c == '\n') break;
      }
      else if(!rc)
      {
         if(n == 1) return;
         else break;
      }
      else return;
   }
   *ptr = 0;
}

short openlogs(int quiet)
{
   int i;

   if(!quiet) { printf("Opening logs..."); fflush(stdout); }
   logfile = malloc(cfg.logcount*sizeof(int));
   for(i=0;i<cfg.logcount;i++) logfile[i] = -1;
   for(i=0;i<cfg.logcount;i++)
      if(cfg.logs[i].path)
      {
         logfile[i] = open(cfg.logs[i].path,O_WRONLY|O_APPEND|O_CREAT,0640);
         if(logfile[i] < 0)
         {
            ftplog("ERR","Cannot open log file '%s': %s",cfg.logs[i].path,
             strerror(errno));
            fprintf(stderr,"can't open '%s': %s\n",cfg.logs[i].path,
             strerror(errno));
            return -1;
         }
      }
   if(!quiet) printf("ok\n"); return 0;
}

void inst_str(char *src)
{
   if (strlen(src) <= strlen(Argv[0])) strcpy(Argv[0], src);
   else
   {
      char *ptr, *ptr2;
      int count;
      ptr = Argv[0] + strlen(Argv[0]);
      for(count = 1;count < Argc; count++) 
      {
        if(Argv[count] == ptr+1)
          ptr += strlen(++ptr);
      }
      if (environ[0] == ptr + 1) 
         for (count = 0; environ[count]; count++)
            if (environ[count] == ptr + 1)
               ptr += strlen(++ptr);
      count = 0;
      for (ptr2 = Argv[0]; ptr2 <= ptr; ptr2++)
      {
         *ptr2 = '\0';
         count++;
      }
      strncpy(Argv[0], src, count);
   }
}

char *getident()
{
   int i,isockfd; struct sockaddr_in ident_addr;
   char recvstring[512], sendstring[15], *ptr = recvstring, *ptr2;

#ifndef HTTPSERV
   if(var.identfetched) return var.ident;
   var.identfetched = 1;
#endif
   if(cfg.debuglevel >= 2) ftplog("AUTH","DEBUG: fetching ident");

   memcpy(&ident_addr, &cli_addr, sizeof(struct sockaddr_in));
   ident_addr.sin_port = htons(IDENTPORT);

   if( (isockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
   {
      ftplog("ERR","socket(): %s",strerror(errno)); return NULL;
   }

   if(connect(isockfd,(struct sockaddr *) &ident_addr, sizeof(struct sockaddr_in)) < 0)
   {
      if(cfg.debuglevel >= 1) ftplog("AUTH","DEBUG: Can't connect to remote identd: %s",strerror(errno));
      close(isockfd); return NULL;
   }
   sprintf(sendstring,"%i,%i\r\n",ntohs(cli_addr.sin_port),cfg.port);
   if(write(isockfd,sendstring,strlen(sendstring)) == -1)
   {
      ftplog("ERR","write(): %s",strerror(errno)); return NULL;
   }
   memset(recvstring,0,512);
   if(read(isockfd,recvstring,511) == -1)
   {
      ftplog("ERR","read(): %s",strerror(errno)); return NULL;
   }
   close(isockfd);
   for(i=0;i<4;i++) if(!(ptr2=strsep(&ptr,":")))
   {
      if(cfg.debuglevel >= 1) ftplog("AUTH","Error parsing ident reply"); return NULL;
   }
   ptr2 = strsep(&ptr2,"\r\n");
   while(*ptr2 && !isgraph(*ptr2))
   {
      *ptr2 = 0; ptr2++;
   }
   ptr = ptr2+strlen(ptr2)-1;
   while(*ptr && !isgraph(*ptr))
   {
      *ptr = 0; ptr--;
   }

   if(strlen(ptr2))
   {
      char *ident = malloc(strlen(ptr2)+1);

      strcpy(ident,ptr2);
      if(cfg.debuglevel >= 1) ftplog("AUTH","DEBUG: remote ident is \"%s\"",ident);
#ifndef HTTPSERV
      var.ident = malloc(strlen(ptr2)+1);
      strcpy(var.ident,ptr2);
#endif
      return ident;
   }
   return NULL;
}

int checkidentip(struct sockaddr_in *chk_addr, int state)
{
   int i, ipok = 0, identok = 0, valid_addrs = 0;
   
   for(i=0;i<USER.identipcount && (!identok || !ipok);i++)
      if(USER.identip[i].flags&state)
      {
         ipok = 0; identok = 0; valid_addrs++;
         if(!USER.identip[i].ip) ipok = 1;
         if(state == ID_DATA || !USER.identip[i].ident) identok = 1;
         if(!ipok)
         {
            if(!fnmatch(USER.identip[i].ip,inet_ntoa(chk_addr->sin_addr),FNM_NOESCAPE | FNM_PERIOD))
            {
               ipok = 1;
               if(cfg.debuglevel >= 1) ftplog("AUTH","DEBUG: IP mask %s matched",USER.identip[i].ip);
            }
            else if(cfg.debuglevel >= 1) ftplog("AUTH","DEBUG: IP mask %s not matched",USER.identip[i].ip);
         }
#ifndef HTTPSERV
         if(!identok && ipok)
         {
            if(getident() && !fnmatch(USER.identip[i].ident,getident(),FNM_NOESCAPE | FNM_PERIOD))
            {
               identok = 1;
               if(cfg.debuglevel >= 1) ftplog("AUTH","DEBUG: ident mask %s matched",USER.identip[i].ident);
            }
            else if(getident() && cfg.debuglevel >= 1) ftplog("AUTH",
             "DEBUG: ident mask %s not matched",USER.identip[i].ident);
         }
#else
         identok = 1;
#endif
         if(ipok && identok)
         {
            if(USER.identip[i].flags&ID_BAN) return 1;
            else return 0;
         }
      }
   if(!valid_addrs) return 0; return -1;
}

void reloadconf(int sig)
{
   int i;
   
#ifndef HTTPSERV
   ftplog("CONF","Reloading configuration file");
#else
   ftplog("CONF","HTTP: Reloading configuration file");
#endif
   for(i=cfg.logcount;i;i--) close(logfile[i-1]);
   dftp_freeconfig(&cfg); dftp_readconfig(&cfg);
   openlogs(1); signal(SIGHUP,reloadconf);
}

void loadplugins()
{
   char plugdirname[PATH_MAX]; struct dirent *dent; DIR *plugindir; void *handle;
   unsigned short order; struct dftp_flag flg; void *symbol;
#ifndef HTTPSERV
   struct dftp_plugininfo pinfo; struct dftp_sitecmd scmd;
   void (*init)(struct dftp_plugininfo);
#endif
   
   /* Construct plugin info structure */

#ifndef HTTPSERV
   memset(&pinfo,0,sizeof(struct dftp_plugininfo));
   pinfo.cfg = &cfg; pinfo.ftplog = ftplog; pinfo.sockfd = sockfd;
   pinfo.var = &var; pinfo.passwdfile = passwdfile; pinfo.groupfile = groupfile;
   pinfo.utmpfile = utmpfile; pinfo.writes = writes; pinfo.cwrites = cwrites;
   pinfo.activity = activity; pinfo.viewfile = viewfile; strcpy(pinfo.ftpdversion,VER);
#endif
   
   /* Load plugins, one by one, function by function */

   sprintf(plugdirname,"%s/%s",cfg.confdir,PLUGINDIR);
   plugindir = opendir(plugdirname);
   if(!plugindir)
   {
      ftplog("ERR","Failed to open plugins directory"); return;
   }
   while((dent = readdir(plugindir)))
   {
      sprintf(plugdirname,"%s/%s/%s",cfg.confdir,PLUGINDIR,dent->d_name);
      if(dent->d_name[0] != '.') handle = dlopen(plugdirname,RTLD_LAZY);
      else handle = NULL;
      if(!handle)
      {
         if(dent->d_name[0] != '.') ftplog("ERR","Could not load plugin %s",dlerror());
      }
      else
      {
        if(cfg.debuglevel >= 2) ftplog("MISC","DEBUG: Loading plugin '%s'",dent->d_name);
#ifndef HTTPSERV
        if((init = dlsym(handle,"initplugin"))) init(pinfo);
        for(order=1;;)
         {
            sprintf(plugdirname,"cmdname%hu",order);
            if(!(scmd.name = dlsym(handle,plugdirname))) break;
            sprintf(plugdirname,"cmdhelp%hu",order);
            if(!(scmd.helptxt = dlsym(handle,plugdirname)))
            {
               ftplog("ERR","Error loading command '%s' - symbol '%s' not found",scmd.name,plugdirname);
               break;
            }
            sprintf(plugdirname,"cmd%hu",order);
            if(!(scmd.cmd = dlsym(handle,plugdirname)))
            {
               ftplog("ERR","Error loading command '%s' - symbol '%s' not found",scmd.name,plugdirname);
               break;
            }
            sprintf(plugdirname,"cmdavail%hu",order);
            if((symbol = dlsym(handle,plugdirname))) scmd.avail = *((char *)symbol);
            else scmd.avail = ALL;
            order++;
            cfg.sitecmds = realloc(cfg.sitecmds,++cfg.sitecmdcount*sizeof(struct dftp_sitecmd));
            memcpy(&cfg.sitecmds[cfg.sitecmdcount-1],&scmd,sizeof(struct dftp_sitecmd));
         }
         if(cfg.debuglevel >= 1) ftplog("MISC","DEBUG: Loaded %i functions of plugin '%s'",order-1,dent->d_name);
#endif
         for(order=1;;)
         {
            sprintf(plugdirname,"flagname%hu",order);
            if(!(flg.name = dlsym(handle,plugdirname))) break;
            sprintf(plugdirname,"flagmark%hu",order);
            if(!(symbol = dlsym(handle,plugdirname)))
            {
               ftplog("ERR","Error loading flag '%s' - symbol '%s' not found",flg.name,plugdirname);
               break;
            }
            else flg.mark = *((char *)symbol);
            order++;
            cfg.flags = realloc(cfg.flags,++cfg.flagcount*sizeof(struct dftp_flag));
            memcpy(&cfg.flags[cfg.flagcount-1],&flg,sizeof(struct dftp_flag));
         }
         if(cfg.debuglevel >= 1) ftplog("MISC","DEBUG: Loaded %i flags of plugin '%s'",order-1,dent->d_name);
      }
   }
}
